# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging
from tkinter import messagebox
import os

# Third Party Imports

# Local Imports
from navigate.tools.multipos_table_tools import (
    sign,
    compute_tiles_from_bounding_box,
    calc_num_tiles,
    update_table,
)
from navigate.controller.sub_controllers.gui import GUIController
from navigate.tools.common_functions import combine_funcs
from navigate.tools.file_functions import save_yaml_file, load_yaml_file
from navigate.config.config import get_navigate_path


# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class TilingWizardController(GUIController):
    """Tiling Wizard Controller

    Controller for tiling wizard parameters.
    Gathers the FOV from the camera settings tab and will
    update when user changes this value.
    Set start/end position buttons will grab the stage
    values for the respective axis when pressed and display in popup
    Number of images we need to acquire with our desired
    percent overlap is calculated and then displayed in third column
    """

    def __init__(self, view, parent_controller):
        """Initialize Tiling Wizard Controller

        Parameters
        ----------
        view : object
            GUI element containing widgets and variables to control.
            Likely tk.Toplevel-derived. In this case tiling_wizard_popup.py
        parent_controller : channels_tab_controller
            The controller that creates the popup/this controller.

        """
        super().__init__(view, parent_controller)

        # Getting widgets and buttons and vars of widgets
        #: dict: Dictionary of widgets in the view
        self.widgets = self.view.get_widgets()

        #: dict: Dictionary of buttons in the view
        self.buttons = self.view.get_buttons()

        #: dict: Dictionary of variables in the view
        self.variables = self.view.get_variables()

        #: int: Default percent overlap between tiles
        self._percent_overlap = 10.0  # default to 10% overlap

        #: dict: flags indicating if all the value are correct to set the table
        self.is_validated = {"x": True, "y": True, "z": True, "f": True}

        # Initialize widgets to previous values
        #: list: List of axes to iterate over
        stage_axes = self.parent_controller.parent_controller.configuration_controller.stage_axes
        self._axes = [axis for axis in stage_axes if axis != "theta"]
        self.load_settings()

        # Ref to widgets in other views
        # (Camera Settings, Stage Control Positions, Stack Acq Settings)
        main_view = self.parent_controller.parent_controller.view
        self.cam_settings_widgets = (
            main_view.settings.camera_settings_tab.camera_roi.get_widgets()
        )
        self.stack_acq_widgets = (
            main_view.settings.channels_tab.stack_acq_frame.get_widgets()
        )
        self.stage_position_vars = (
            main_view.settings.stage_control_tab.position_frame.get_variables()
        )
        self.multipoint_table = (
            main_view.settings.multiposition_tab.multipoint_list.get_table()
        )

        # Setting/Tracing Percent Overlay, also handled in update_overlap
        self.variables["percent_overlap"].trace_add(
            "write", lambda *args: self.update_overlap()
        )

        # Trace cam_settings FOV to catch user changes
        # FOV change handled in update_fov
        self.cam_settings_widgets["FOV_X"].get_variable().trace_add(
            "write", lambda *args: self.update_fov("y")
        )
        self.cam_settings_widgets["FOV_Y"].get_variable().trace_add(
            "write", lambda *args: self.update_fov("x")
        )
        # primary z/f
        self.primary_z_axis = self.stack_acq_widgets["z_device"].get().split(" - ")[1]
        self.primary_f_axis = self.stack_acq_widgets["f_device"].get().split(" - ")[1]
        self.stack_acq_widgets["z_device"].get_variable().trace_add(
            "write", lambda *args: self.update_fov("z_device")
        )
        self.stack_acq_widgets["f_device"].get_variable().trace_add(
            "write", lambda *args: self.update_fov("f_device")
        )
        self.stack_acq_widgets["start_position"].get_variable().trace_add(
            "write", lambda *args: self.update_fov("z")
        )
        self.stack_acq_widgets["end_position"].get_variable().trace_add(
            "write", lambda *args: self.update_fov("z")
        )
        self.stack_acq_widgets["start_focus"].get_variable().trace_add(
            "write", lambda *args: self.update_fov("f")
        )
        self.stack_acq_widgets["end_focus"].get_variable().trace_add(
            "write", lambda *args: self.update_fov("f")
        )

        # Calculate distances
        for axis in self._axes:
            self.variables[f"{axis}_start"].trace_add(
                "write", lambda *args, axis=axis: self.calculate_distance(axis)
            )
            self.variables[f"{axis}_end"].trace_add(
                "write", lambda *args, axis=axis: self.calculate_distance(axis)
            )
            # Bind FOV changes
            self.variables[f"{axis}_fov"].trace_add(
                "write", lambda *args, axis=axis: self.calculate_tiles(axis)
            )

            # Calculating Number of Tiles traces
            self.variables[f"{axis}_dist"].trace_add(
                "write", lambda *args, axis=axis: self.calculate_tiles(axis)
            )

        # Populate Table trace
        self.buttons["set_table"].configure(command=self.set_table)

        for ax in self._axes:
            # Start/End buttons
            self.buttons[f"{ax}_start"].configure(
                command=self.position_handler(ax, "start")
            )
            self.buttons[f"{ax}_end"].configure(
                command=self.position_handler(ax, "end")
            )

            # Calculating total tile traces
            self.variables[f"{ax}_tiles"].trace_add(
                "write", lambda *args: self.update_total_tiles()
            )

            # Update widgets to current values in other views
            self.update_fov(ax)

        # Properly Closing Popup with parent controller
        self.view.popup.protocol(
            "WM_DELETE_WINDOW",
            self.close_window,
        )

        self.view.popup.bind("<Escape>", self.close_window)

    def load_settings(self):
        """Load positions from yaml file"""

        # Load positions from yaml file
        load_path = os.path.join(
            get_navigate_path(), "config", "tiling_wizard_settings.yaml"
        )

        if os.path.exists(load_path):
            positions = load_yaml_file(load_path)
            # for key, value in positions.items():
            #     self.variables[key].set(value)
        else:
            positions = {}

        self.variables["percent_overlap"].set(self._percent_overlap)
        self.variables["total_tiles"].set(1)
        for ax in self._axes:
            self.variables[f"{ax}_start"].set(positions.get(f"{ax}_start", 0.0))
            self.variables[f"{ax}_end"].set(positions.get(f"{ax}_end", 0.0))
            self.variables[f"{ax}_dist"].set(positions.get(f"{ax}_dist", 0.0))
            self.variables[f"{ax}_fov"].set(positions.get(f"{ax}_fov", 0.0))
            self.variables[f"{ax}_tiles"].set(positions.get(f"{ax}_tiles", 1))

    def close_window(self, *args) -> None:
        """Save multiposition information and close the tiling wizard popup

        Save multiposition information to yaml file, and close tiling wizard.
        """
        positions = {key: var.get() for key, var in self.variables.items()}
        save_yaml_file(
            file_directory=os.path.join(get_navigate_path(), "config"),
            content_dict=positions,
            filename="tiling_wizard_settings.yaml",
        )

        # Close the popup
        self.view.popup.dismiss()
        delattr(self.parent_controller, "tiling_wizard_controller")

    def set_table(self):
        """Set the multipoint table to the values in the tiling wizard

        Sets multiposition table with values from tiling wizard after
        populate Multiposition Table button is pressed
        Compute grid will return a list of all position combinations.
        This list is then converted to a
        pandas dataframe which is then set as the new table data.
        The table is then redrawn.
        """

        def sort_vars(a, b):
            """Sort two variables from low to high

            Parameters
            ----------
            a : float
                First variable
            b : float
                Second variable

            Returns
            -------
            a, b : float
                Sorted variables
            """
            if a > b:
                return b, a
            return a, b
    
        if False in self.is_validated.values():
            messagebox.showwarning(
                title="Navigate",
                message="Can't calculate positions, "
                "please make sure all FOV Dists are correct!",
            )
            return
        
        tiling_setting = {}
        for axis in self._axes:
            start_pos = float(self.variables[f"{axis}_start"].get())
            stop_pos = float(self.variables[f"{axis}_end"].get())
            tiles = int(self.variables[f"{axis}_tiles"].get())
            fov = float(self.variables[f"{axis}_fov"].get())

            if axis == self.primary_z_axis:
                start_pos -= float(
                    self.stack_acq_widgets["start_position"].get()
                )
                stop_pos -= float(
                    self.stack_acq_widgets["end_position"].get()
                )
            elif axis == self.primary_f_axis:
                start_pos -= float(
                    self.stack_acq_widgets["start_focus"].get()
                )
                stop_pos -= float(
                    self.stack_acq_widgets["end_focus"].get()
                )

            start_pos, stop_pos = sort_vars(start_pos, stop_pos)
            tiling_setting[f"{axis}_start"] = start_pos
            # tiling_setting[f"{axis}_stop"] = stop_pos
            tiling_setting[f"{axis}_tiles"] = tiles
            tiling_setting[f"{axis}_length"] = fov

        tiling_setting["theta_start"] = float(self.stage_position_vars["theta"].get())
        tiling_setting["theta_tiles"] = 1
        tiling_setting["theta_length"] = 0

        overlap = float(self._percent_overlap) / 100
        columns, table_values = compute_tiles_from_bounding_box(
            overlap=overlap,
            **tiling_setting
        )

        update_table(self.multipoint_table, table_values, columns)

        # If we have additional axes, create self.d{axis} for each
        # additional axis, to ensure we keep track of the step size
        config = self.parent_controller.parent_controller.configuration
        microscope_name = config["experiment"]["MicroscopeState"]["microscope_name"]
        scope = config["configuration"]["microscopes"][microscope_name]
        coupled_axes = scope["stage"].get("coupled_axes", None)
        if coupled_axes is not None:
            for follower in coupled_axes.values():
                config["experiment"]["MicroscopeState"][
                    f"{follower.lower()}_step_size"
                ] = self.variables[f"{follower.lower()}_fov"].get()

    def update_total_tiles(self):
        """Update the total number of tiles in the tiling wizard

        Sums the tiles for each axis in the tiling wizard.
        Will update when any axis has a tile amount change.
        """

        total_tiles = 1
        for ax in self._axes:
            total_tiles *= float(self.variables[f"{ax}_tiles"].get())

        self.variables["total_tiles"].set(total_tiles)

    def calculate_tiles(self, axis=None):
        """Calculate the number of tiles for a given axis

        Calculates the number of tiles of the acquisition for
        each axis or an individual axis
        Num of Tiles = dist - (overlay * FOV)  /  FOV * (1 - overlay)
        (D-OF)/(F-OF) = N

        Parameters
        ----------
        axis : str
            x, y, z, f axis of stage to calculate.
        """

        if axis not in self._axes + [None]:
            logger.warning(
                f"Controller - Tiling Wizard - Unknown axis {axis}, "
                "skipping calculate_tiles()."
            )
            return

        if axis is not None:
            if not isinstance(axis, list):
                axis = [axis]
        else:
            axis = self._axes

        overlay = float(self._percent_overlap) / 100

        for ax in axis:
            self.is_validated[ax] = True
            fov_value = self.variables[f"{ax}_fov"].get()
            if not fov_value or "inf" in fov_value:
                self.is_validated[ax] = False
                return
            try:
                dist = abs(float(self.variables[f"{ax}_dist"].get()))  # um
                fov = abs(float(self.variables[f"{ax}_fov"].get()))  # um

                if ax.lower() == "x" or ax.lower() == "y":
                    # + fov because distance is center of the fov to center of
                    # the fov and so we are covering a distance that is 2 *
                    # 1/2 * fov larger than dist
                    dist += fov

                num_tiles = calc_num_tiles(dist, overlay, fov)

                self.variables[f"{ax}_tiles"].set(num_tiles)
            except ValueError as e:
                self.is_validated[ax] = False
                logger.warning(f"Controller - Tiling Wizard - {e}")

    def calculate_distance(self, axis):
        """Calculate the distance for a given axis

        This function will calculate the distance for a given
        axis of the stage when the start or end position is changed
        via the Set buttons

        Parameters
        ----------
        axis : str
            x, y, z axis of stage to calculate
        """

        start = float(self.variables[axis + "_start"].get())
        end = float(self.variables[axis + "_end"].get())
        dist = abs(end - start)
        self.variables[axis + "_dist"].set(dist)

    def update_overlap(self):
        """Update the overlay percentage for the tiling wizard

        Updates percent overlay when a user changes the widget in the popup.
        This value is used for backend calculations.
        The number of tiles will then be recalculated
        """

        try:
            self._percent_overlap = float(self.variables["percent_overlap"].get())
            self.calculate_tiles()
        except ValueError:
            # most likely an empty string was passed
            pass

    def position_handler(self, axis, start_end):
        """Set the start or end position for a given axis

        When the Set [axis] Start/End button is pressed then the
        stage position is polled from the stage controller

        Parameters
        ----------
        axis : str
            x, y, z, f axis that corresponds to stage axis
        start_end : str
            start or end will signify which spinbox gets updated upon button press

        Returns
        -------
        handler : func
            Function for setting positional spinbox based on parameters passed in
        """

        def handler():
            # Force us to get the current stage positions from the stage
            self.parent_controller.parent_controller.execute("stop_stage")

            def set_bounds(axis, start_end):
                # Now set the bounds
                pos = self.stage_position_vars[axis].get()
                self.widgets[axis + "_" + start_end].widget.set(pos)
                # if axis == "z":
                #     setattr(self, f"_f_{start_end}",
                #             self.stage_position_vars["f"].get())

            self.parent_controller.parent_controller.view.after(
                250, lambda: set_bounds(axis, start_end)
            )

        return handler

    def update_fov(self, axis=None):
        """Update the FOV for the tiling wizard

        Grabs the updated FOV if changed by user,
        will recalculate num of tiles for each axis after

        Parameters
        ----------
        axis : str
            Axis
        """

        if axis is None:
            axes = self._axes
        elif isinstance(axis, str):
            if axis == "z_device":
                # get the new primary z axis
                primary_z = self.stack_acq_widgets["z_device"].get().split(" - ")[1]
                if self.primary_z_axis != primary_z:
                    self.variables[f"{primary_z}_fov"].set(self.variables[f"{self.primary_z_axis}_fov"].get())
                    self.variables[f"{self.primary_z_axis}_fov"].set(0)
                    self.primary_z_axis = primary_z
                return
            elif axis == "f_device":
                # get the new primary f axis
                primary_f = self.stack_acq_widgets["f_device"].get().split(" - ")[1]
                if self.primary_f_axis != primary_f:
                    self.variables[f"{primary_f}_fov"].set(self.variables[f"{self.primary_f_axis}_fov"].get())
                    self.variables[f"{self.primary_f_axis}_fov"].set(0)
                    self.primary_f_axis = primary_f
                return
            else:
                axes = [axis]

        for ax in axes:
            try:
                # Calculate signed fov
                if ax == "y":
                    y = float(self.cam_settings_widgets["FOV_X"].get()) * sign(
                        float(self.variables["x_end"].get())
                        - float(self.variables["x_start"].get())
                    )
                    axis = "y"
                elif ax == "x":
                    x = float(self.cam_settings_widgets["FOV_Y"].get()) * sign(
                        float(self.variables["y_end"].get())
                        - float(self.variables["y_start"].get())
                    )
                    axis = "x"
                elif ax == "z":
                    z = float(self.stack_acq_widgets["end_position"].get()) - float(
                        self.stack_acq_widgets["start_position"].get()
                    )
                    axis = self.primary_z_axis
                elif ax == "f":
                    f = float(self.stack_acq_widgets["end_focus"].get()) - float(
                        self.stack_acq_widgets["start_focus"].get()
                    )
                    axis = self.primary_f_axis

                # for ax in self._axes:
                # self._fov[ax] = locals().get(ax)
                self.variables[f"{axis}_fov"].set(
                    abs(locals().get(ax))
                )  # abs(self._fov[ax]))

                self.calculate_tiles(axis)
            except (TypeError, ValueError) as e:
                logger.debug(
                    f"Controller - Tiling Wizard - Caught ValueError: {e}. "
                    f"Declining to update {ax} FOV."
                )
                pass

    def showup(self):
        """Show the tiling wizard

        Brings popup window to front of screen
        """
        self.view.popup.deiconify()
        self.view.popup.attributes("-topmost", 1)



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard Library Imports
import logging
import tkinter as tk
from tkinter import messagebox
from typing import Dict, Any, Iterable
import re

# Third Party Imports

# Local Imports
from navigate.controller.sub_controllers.gui import GUIController
from navigate.tools.file_functions import create_save_path
from navigate.view.popups.acquire_popup import AcquirePopUp
from navigate.view.main_window_content.acquire_notebook import AcquireBar
from navigate.config import update_config_dict

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class AcquireBarController(GUIController):
    """Acquire Bar Controller."""

    def __init__(self, view: AcquireBar, parent_controller: Any) -> None:
        """Initialize the Acquire Bar Controller.

        Parameters
        ----------
        view : AcquireBar
            Instance of the View.
        parent_controller : Any
            Instance of the Main Controller.
        """
        super().__init__(view, parent_controller)

        #: AcquirePopUp: Instance of the popup window.
        self.acquire_pop = None

        #: dict: Saving settings.
        self.saving_settings = None

        #: str: Acquisition image mode.
        self.mode = "live"

        #: bool: Whether the image will be saved.
        self.is_save = False

        #: bool: Whether the microscope is acquiring.
        self.is_acquiring = False

        #: dict: Dictionary of different operating modes.
        self.mode_dict = {
            "Continuous Scan": "live",
            "Z-Stack": "z-stack",
            "Single Acquisition": "single",
            "Customized": "customized",
        }

        self.view.pull_down["values"] = list(self.mode_dict.keys())
        self.view.pull_down.current(0)
        self.view.pull_down.state(["!disabled", "readonly"])

        # gui event bind
        self.view.acquire_btn.config(command=self.launch_popup_window)
        self.view.pull_down.bind("<<ComboboxSelected>>", self.update_microscope_mode)
        self.view.exit_btn.config(command=self.exit_program)

        # framerate information.
        self.framerate = 0

        #: dict: The ProxyDict for the BDV configuration.
        self.bdv_configuration = self.parent_controller.configuration["experiment"].get(
            "BDVParameters", None
        )

        #: dict: BDV Widgets. Keys are primary widgets, values are lists.
        self.bdv_widgets = {
            "shear_data": ["shear_dimension", "shear_angle"],
            "rotate_data": ["rotate_angle_x", "rotate_angle_y", "rotate_angle_z"],
            "down_sample_data": ["lateral_down_sample", "axial_down_sample"],
        }

    def progress_bar(
        self,
        images_received: int,
        microscope_state: Dict[str, Any],
        mode: str,
        stop=False,
    ):
        """Update progress bars.

        Parameters
        ----------
        microscope_state : dict
            State of the microscope.  Positions, channels, z-steps, etc.
        mode : str
            Imaging mode.  'live', 'z-stack', ...
        images_received : int
            Number of images received in the controller.
        stop : bool
            Stop flag to set back to 0.
        """

        if images_received == 0:
            if mode == "live" or mode == "customized":
                # Set to Indeterminate mode.
                self.view.CurAcq["mode"] = "indeterminate"
                self.view.OvrAcq["mode"] = "indeterminate"
            else:
                # Set to Determinate mode and initialize at zero.
                # stack_index = 0
                self.view.CurAcq["mode"] = "determinate"
                self.view.OvrAcq["mode"] = "determinate"
                self.view.CurAcq["value"] = 0
                self.view.OvrAcq["value"] = 0

        # Calculate the number of images anticipated.
        number_of_channels = 0
        for channel in microscope_state["channels"].keys():
            if microscope_state["channels"][channel]["is_selected"] is True:
                number_of_channels += 1

        # Time-lapse acquisition
        number_of_timepoints = int(microscope_state["timepoints"])

        # Multi-Position Acquisition
        if microscope_state["is_multiposition"] is False:
            number_of_positions = 1
        else:
            # The first row is a header that describes stage axis designation.
            number_of_positions = len(
                self.parent_controller.configuration["multi_positions"]
            ) - 1

        if mode == "single":
            number_of_slices = 1
        elif mode == "z-stack":
            number_of_slices = microscope_state["number_z_steps"]
        else:
            number_of_slices = 1

        top_anticipated_images = number_of_slices
        bottom_anticipated_images = (
            number_of_channels
            * number_of_slices
            * number_of_timepoints
            * number_of_positions
        )

        if images_received > 0:
            # Update progress bars according to imaging mode.
            if stop is False:
                if mode == "z-stack" or mode == "single":
                    # Calculate the number of images remaining.
                    # Time is estimated from the framerate, which includes stage
                    # movement time inherently.
                    try:
                        images_remaining = bottom_anticipated_images - images_received
                        seconds_left = images_remaining / self.framerate
                        self.update_progress_label(seconds_left)
                    except ZeroDivisionError:
                        pass
                else:
                    self.view.CurAcq.start()
                    self.view.OvrAcq.start()
                    self.view.total_acquisition_label.config(text="--:--:--")

                if mode == "z-stack":
                    top_percent_complete = 100 * (
                        images_received / top_anticipated_images
                    )

                    self.view.CurAcq["value"] = (
                        top_percent_complete % 100
                        if (top_percent_complete > 100.0)
                        else top_percent_complete
                    )

                    bottom_anticipated_images = 100 * (
                        images_received / bottom_anticipated_images
                    )
                    self.view.OvrAcq["value"] = bottom_anticipated_images

                elif mode == "single":
                    top_percent_complete = 100 * (
                        images_received / top_anticipated_images
                    )
                    self.view.CurAcq["value"] = top_percent_complete
                    self.view.OvrAcq["value"] = top_percent_complete

                else:
                    self.view.CurAcq.start()
                    self.view.OvrAcq.start()

            elif stop is True:
                self.update_progress_label(seconds_left=0)
                self.stop_progress_bar()

    def stop_progress_bar(self) -> None:
        """Stop moving the continuous progress bar."""
        self.view.CurAcq.stop()
        self.view.OvrAcq.stop()

    def update_progress_label(self, seconds_left: int) -> None:
        """Update the progress label in the Acquire Bar.

        Formatted time is in HH:MM:SS.

        Parameters
        ----------
        seconds_left : int
            Seconds left in the acquisition.
        """
        hours, remainder = divmod(seconds_left, 3600)
        minutes, seconds = divmod(remainder, 60)
        self.view.total_acquisition_label.config(
            text=f"{int(hours):02}" f":{int(minutes):02}" f":{int(seconds):02}"
        )

    def set_mode(self, mode: str) -> None:
        """Set imaging mode.

        Parameters
        ----------
        mode: str
            Mode could be: 'live', 'z-stack', 'single', 'projection',
        """
        # update pull down combobox
        reverse_dict = dict(map(lambda v: (v[1], v[0]), self.mode_dict.items()))
        if mode not in reverse_dict:
            mode = list(reverse_dict.keys())[0]
        self.mode = mode
        self.view.pull_down.set(reverse_dict[mode])
        self.parent_controller.configuration["experiment"]["MicroscopeState"][
            "image_mode"
        ] = mode
        self.show_verbose_info("Image mode is set to", mode)

    def get_mode(self) -> str:
        """Get the current imaging mode.

        Returns
        -------
        mode : str
            Current imaging mode.
        """
        return self.mode

    def add_mode(self, mode: str) -> None:
        """Add a new mode to the mode dictionary.

        Parameters
        ----------
        mode : str
            Mode to add to the dictionary.
        """
        if mode not in self.mode_dict:
            self.mode_dict[mode] = mode
            self.view.pull_down["values"] = list(self.mode_dict.keys())

    def stop_acquire(self) -> None:
        """Stop the acquisition.

        Stop the progress bar, set the acquire button back to "Acquire", place pull
        down menu in non-disabled format.
        """
        self.stop_progress_bar()
        self.view.acquire_btn.configure(text="Acquire")
        self.view.acquire_btn.configure(state="normal")
        self.view.pull_down.state(["!disabled", "readonly"])
        self.is_acquiring = False

    def set_save_option(self, is_save: bool) -> None:
        """Set whether the image will be saved.

        Parameters
        ----------
        is_save : bool
            True if we will save the data.  False if we will not.
        """
        self.is_save = is_save
        self.parent_controller.configuration["experiment"]["MicroscopeState"][
            "is_save"
        ] = is_save
        self.show_verbose_info("set save data option:", is_save)

    def launch_popup_window(self) -> None:
        """Launch the Save Dialog Popup Window

        The popup window should only be launched if the microscope is set to save the
        data, except the continuous acquisition mode. The popup window provides the
        user with the opportunity to fill in fields that describe the experiment and
        also dictate the save path of the data in a standardized format.
        """
        if self.is_acquiring and self.view.acquire_btn["text"] == "Acquire":
            return

        if not self.is_acquiring and self.view.acquire_btn["text"] == "Stop":
            return

        if self.view.acquire_btn["text"] == "Stop":
            # tell the controller to stop acquire (continuous mode)
            self.view.acquire_btn.configure(state="disabled")
            self.parent_controller.execute("stop_acquire")

        elif self.is_save and self.mode != "live":
            self.acquire_pop = AcquirePopUp(self.view)

            buttons = self.acquire_pop.get_buttons()
            widgets = self.acquire_pop.get_widgets()

            # Configure the button callbacks on the popup window
            buttons["Cancel"].config(command=lambda: self.acquire_pop.popup.dismiss())
            buttons["Done"].config(
                command=lambda: self.launch_acquisition(self.acquire_pop)
            )
            self.acquire_pop.popup.bind(
                "<Escape>", lambda e: self.acquire_pop.popup.dismiss()
            )

            # Configure drop down callbacks, will update save settings when file type is
            # changed
            file_type = widgets["file_type"].get_variable()
            file_type.trace_add("write", lambda *args: self.update_file_type(file_type))

            self.set_bdv_settings()

            for widget in self.bdv_widgets.keys():
                # Traces for the main widgets
                self.acquire_pop.tab_frame.inputs[widget].get_variable().trace_add(
                    "write",
                    lambda *args, main_widget=widget, dep_widget=self.bdv_widgets[
                        widget
                    ]: self.toggle_bdv_widgets(main_widget, dep_widget),
                )

            for k, v in self.saving_settings.items():
                if widgets.get(k, None):
                    widgets[k].set(v)
            self.acquire_pop.tab_frame.inputs["misc"].insert(
                "1.0", self.saving_settings.get("misc", "")
            )

        else:
            self.is_acquiring = True
            self.view.acquire_btn.configure(state="disabled")
            self.view.pull_down.state(["disabled", "readonly"])
            self.parent_controller.execute("acquire")

    def set_bdv_settings(self) -> None:
        """Set the BDV Settings from the configuration file.

        If the settings are not in the configuration file, they will be added.
        """
        update_config = False

        if self.bdv_configuration is None:
            update_config = True
            self.bdv_configuration = {}

        # Shear Parameters
        if self.bdv_configuration.get("shear", None) is None:
            update_config = True
            self.bdv_configuration["shear"] = {
                "shear_data": False,
                "shear_dimension": "YZ",
                "shear_angle": 0,
            }
        self.acquire_pop.tab_frame.inputs["shear_data"].set(
            self.bdv_configuration["shear"].get("shear_data", False)
        )
        self.acquire_pop.tab_frame.inputs["shear_dimension"].set(
            self.bdv_configuration["shear"].get("shear_dimension", "YZ")
        )
        self.acquire_pop.tab_frame.inputs["shear_angle"].set(
            self.bdv_configuration["shear"].get("shear_angle", 0)
        )

        # Rotation Parameters
        if self.bdv_configuration.get("rotate", None) is None:
            update_config = True
            self.bdv_configuration["rotate"] = {
                "rotate_data": False,
                "X": 0,
                "Y": 0,
                "Z": 0,
            }
        self.acquire_pop.tab_frame.inputs["rotate_data"].set(
            self.bdv_configuration["rotate"].get("rotate_data", False)
        )
        self.acquire_pop.tab_frame.inputs["rotate_angle_x"].set(
            self.bdv_configuration["rotate"].get("X", 0)
        )
        self.acquire_pop.tab_frame.inputs["rotate_angle_y"].set(
            self.bdv_configuration["rotate"].get("Y", 0)
        )
        self.acquire_pop.tab_frame.inputs["rotate_angle_z"].set(
            self.bdv_configuration["rotate"].get("Z", 0)
        )

        # Down Sample Parameters
        if self.bdv_configuration.get("down_sample", None) is None:
            update_config = True
            self.bdv_configuration["down_sample"] = {
                "down_sample": False,
                "lateral_down_sample": 1,
                "axial_down_sample": 1,
            }
        self.acquire_pop.tab_frame.inputs["down_sample_data"].set(
            self.bdv_configuration["down_sample"].get("down_sample", False)
        )
        self.acquire_pop.tab_frame.inputs["lateral_down_sample"].set(
            str(self.bdv_configuration["down_sample"].get("lateral_down_sample", 1))
            + "x"
        )
        self.acquire_pop.tab_frame.inputs["axial_down_sample"].set(
            str(self.bdv_configuration["down_sample"].get("axial_down_sample", 1)) + "x"
        )

        if update_config:
            update_config_dict(
                self.parent_controller.manager,
                self.parent_controller.configuration["experiment"],
                "BDVParameters",
                self.bdv_configuration,
            )
            self.bdv_configuration = self.parent_controller.configuration["experiment"][
                "BDVParameters"
            ]

        # Set the state of the dependent widgets
        for widget in self.bdv_widgets.keys():
            self.toggle_bdv_widgets(widget, self.bdv_widgets[widget])

    def update_bdv_settings(self) -> None:
        """Update the BDV settings."""
        if self.bdv_configuration is not None:
            # Set the widget variables according to the BDV configuration
            # Shear Parameters
            self.bdv_configuration["shear"]["shear_data"] = bool(
                (
                    update := self.acquire_pop.tab_frame.inputs[
                        "shear_data"
                    ].variable.get()
                )
            )

            if update:
                self.bdv_configuration["shear"]["shear_dimension"] = str(
                    self.acquire_pop.tab_frame.inputs["shear_dimension"].variable.get()
                )
                self.bdv_configuration["shear"]["shear_angle"] = float(
                    self.acquire_pop.tab_frame.inputs["shear_angle"].variable.get()
                )

            # Rotation Parameters
            self.bdv_configuration["rotate"]["rotate_data"] = bool(
                (
                    update := self.acquire_pop.tab_frame.inputs[
                        "rotate_data"
                    ].variable.get()
                )
            )

            if update:
                self.bdv_configuration["rotate"]["X"] = float(
                    self.acquire_pop.tab_frame.inputs["rotate_angle_x"].variable.get()
                )
                self.bdv_configuration["rotate"]["Y"] = float(
                    self.acquire_pop.tab_frame.inputs["rotate_angle_y"].variable.get()
                )
                self.bdv_configuration["rotate"]["Z"] = float(
                    self.acquire_pop.tab_frame.inputs["rotate_angle_z"].variable.get()
                )

            # Down Sample Parameters
            self.bdv_configuration["down_sample"]["down_sample"] = bool(
                (
                    update := self.acquire_pop.tab_frame.inputs[
                        "down_sample_data"
                    ].variable.get()
                )
            )

            if update:
                self.bdv_configuration["down_sample"]["lateral_down_sample"] = int(
                    self.acquire_pop.tab_frame.inputs[
                        "lateral_down_sample"
                    ].variable.get()[:-1]
                )

                self.bdv_configuration["down_sample"]["axial_down_sample"] = int(
                    self.acquire_pop.tab_frame.inputs[
                        "axial_down_sample"
                    ].variable.get()[:-1]
                )

    def toggle_bdv_widgets(self, main_widget: str, dependent_widgets: list) -> None:
        """Toggles the state of the dependent widgets.

        Parameters
        ----------
        main_widget : str
            The main widget that is being toggled.
        dependent_widgets : list
            A list of dependent widgets that will be toggled.
        """
        state = self.acquire_pop.tab_frame.inputs[main_widget].get_variable().get()
        for widget in dependent_widgets:
            self.acquire_pop.tab_frame.inputs[widget].widget.config(
                state="readonly" if state else "disabled"
            )

    def update_microscope_mode(self, *args: Iterable) -> None:
        """Gets the state of the pull-down menu and tells the central controller

        Will additionally call functions to disable and enable widgets based on mode

        Parameters
        ----------
        args : str
            Imaging Mode.
        """
        self.mode = self.mode_dict[self.view.pull_down.get()]
        self.show_verbose_info("The Microscope State is now:", self.get_mode())
        # acquire_bar_controller - update image mode
        self.parent_controller.configuration["experiment"]["MicroscopeState"][
            "image_mode"
        ] = self.mode
        # Update state status of other widgets in the GUI based on what mode is set
        self.parent_controller.channels_tab_controller.set_mode("stop")

    def update_file_type(self, file_type: tk.StringVar) -> None:
        """Updates the file type when the drop-down in save dialog is changed.

        Parameters
        ----------
        file_type : tk.StringVar
            File type.
        """
        self.saving_settings["file_type"] = file_type.get()

    def launch_acquisition(self, popup_window: AcquirePopUp) -> None:
        """Launch the Acquisition.

        Once the popup window has been filled out, we first create the save path using
        the create_save_path function.
        This automatically removes spaces and replaces them with underscores.
        Then it makes the directory.

        Parameters
        ----------
        popup_window : AcquirePopUp
            Instance of the popup save dialog.
        """
        # Get the BDV Settings.

        # update saving settings according to user's input
        self.update_experiment_values(popup_window)
        self.update_bdv_settings()

        entry_names = [
            "user",
            "tissue",
            "celltype",
            "label",
            "prefix",
        ]

        for name in entry_names:
            if not self.is_valid_string(self.saving_settings[name]):
                messagebox.showwarning(
                    title="Invalid Entry",
                    message="Only alphanumeric characters, hyphens, "
                    "and underscores are allowed. \n",
                    parent=popup_window.popup,
                )
                return

        # Verify user's input is non-zero.
        is_valid = (
            self.saving_settings["user"]
            and self.saving_settings["tissue"]
            and self.saving_settings["celltype"]
            and self.saving_settings["label"]
        )

        if is_valid:
            # Verify that the path is valid.
            try:
                file_directory = create_save_path(self.saving_settings)
            except Exception:
                messagebox.showwarning(
                    title="Directory Not Found.",
                    message="The directory specified is invalid. \n"
                    "This commonly occurs when the Root Directory is "
                    "incorrect. Please double-check and try again.",
                    parent=popup_window.popup,
                )
                return

            self.is_acquiring = True
            self.view.acquire_btn.configure(state="disabled")
            popup_window.popup.dismiss()
            self.parent_controller.execute("acquire_and_save", file_directory)

    def exit_program(self) -> None:
        """Exit Button to close the program."""
        if messagebox.askyesno("Exit", "Are you sure?"):
            self.show_verbose_info("Exiting Program")
            # call the central controller to stop all the threads
            self.parent_controller.execute("exit")

    def populate_experiment_values(self) -> None:
        """Populate the experiment values from the config file."""
        #: dict: Saving settings.
        self.saving_settings = self.parent_controller.configuration["experiment"][
            "Saving"
        ]
        self.saving_settings["date"] = str(self.saving_settings["date"])
        mode = self.parent_controller.configuration["experiment"]["MicroscopeState"][
            "image_mode"
        ]
        self.set_mode(mode)
        is_save = self.parent_controller.configuration["experiment"]["MicroscopeState"][
            "is_save"
        ]
        self.set_save_option(is_save)

    def update_experiment_values(self, popup_window: AcquirePopUp) -> None:
        """Gets the entries from the popup save dialog and overwrites the
        saving_settings dictionary.

        Parameters
        ----------
        popup_window : AcquirePopUp
            Instance of the popup save dialog.
        """
        popup_vals = popup_window.get_variables()
        for name in popup_vals:
            # remove leading and tailing whitespaces
            self.saving_settings[name] = popup_vals[name].strip()
        self.saving_settings["misc"] = self.acquire_pop.tab_frame.inputs["misc"].get(
            "1.0", "end-1c"
        )

    @staticmethod
    def is_valid_string(string: str) -> bool:
        """Check if the string is valid.

        Parameters
        ----------
        string : str
            String to check.

        Returns
        -------
        bool
            True if the string is valid.
        """
        pattern = r"^[\w\-\ ]+$"
        return bool(re.match(pattern, string))



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


#  Standard Library Imports
import logging
from typing import Any, Dict

# Third Party Imports

# Local Imports
from navigate.model.devices.remote_focus.base import RemoteFocusBase
from navigate.model.devices.device_types import SerialDevice
from navigate.model.devices.APIs.asi.asi_tiger_controller import TigerController
from navigate.tools.decorators import log_initialization

# # Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class ASIRemoteFocus(RemoteFocusBase , SerialDevice):
    """ASIRemoteFocus Class - Analog control of the remote focus device."""

    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        *args,
        **kwargs,
    ) -> None:
        """Initialize the ASIRemoteFocus class.

        Parameters
        ----------
        microscope_name : str
            The microscope name.
        device_connection : Any
            The device connection object.
        configuration : Dict[str, Any]
            The configuration dictionary.

        """

        super().__init__(microscope_name, device_connection, configuration)

        #: Any: Device connection object.
        self.remote_focus = device_connection

        #: str: Output axis on Tiger Controller
        self.axis = self.device_config["hardware"]["axis"]

    @classmethod
    def connect(cls, port, baudrate=115200, timeout=0.25):
        """Build ASILaser Serial Port connection

        Parameters
        ----------
        port : str
            Port for communicating with the filter wheel, e.g., COM1.
        baudrate : int
            Baud rate for communicating with the filter wheel, default is 115200.
        timeout : float
            Timeout for communicating with the filter wheel, default is 0.25.

        Returns
        -------
        tiger_controller : TigerController
            ASI Tiger Controller object.
        """
        # wait until ASI device is ready
        tiger_controller = TigerController(port, baudrate)
        tiger_controller.connect_to_serial()
        if not tiger_controller.is_open():
            logger.error("ASI stage connection failed.")
            raise Exception("ASI stage connection failed.")
        return tiger_controller

    def adjust(self, exposure_times, sweep_times, offset=None):
        """Adjust the waveform.

        This method adjusts the waveform parameters.
        Based on the sensor mode and readout direction, either the triangle or ramp function will be called. 

        Parameters
        ----------
        exposure_times : dict
            Dictionary of exposure times for each selected channel
        sweep_times : dict
            Dictionary of sweep times for each selected channel
        offset : float, optional
            Offset value for the remote focus waveform, by default None
        """

        # to determine if the waveform has to be triangular
        sensor_mode = self.configuration["experiment"]["CameraParameters"][
            self.microscope_name
        ]["sensor_mode"]
        readout_direction = self.configuration["experiment"]["CameraParameters"][
            self.microscope_name
        ]["readout_direction"]

        microscope_state = self.configuration["experiment"]["MicroscopeState"]
        waveform_constants = self.configuration["waveform_constants"]
        imaging_mode = microscope_state["microscope_name"]
        zoom = microscope_state["zoom"]

        for channel_key in microscope_state["channels"].keys():
            # channel includes 'is_selected', 'laser', 'filter', 'camera_exposure'...
            channel = microscope_state["channels"][channel_key]

            # Only proceed if it is enabled in the GUI
            if channel["is_selected"] is True:

                # Get the Waveform Parameters - Assumes ETL Delay < Camera Delay.
                # Should Assert.
                laser = channel["laser"]
                exposure_time = exposure_times[channel_key]
                self.sweep_time = sweep_times[channel_key]

                # Remote Focus Parameters
                # Validation for when user puts a '-' or '.' in spinbox
                temp = waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                    laser
                ]["amplitude"]
                if temp == "-" or temp == ".":
                    waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                        laser
                    ]["amplitude"] = "1000"
                
                amplitude = float(
                    waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                        laser
                    ]["amplitude"]
                )

                # Validation for when user puts a '-' in spinbox
                temp = waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                    laser
                ]["offset"]
                if temp == "-" or temp == ".":
                    waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                        laser
                    ]["offset"] = "0"

                remote_focus_offset = float(
                    waveform_constants["remote_focus_constants"][imaging_mode][zoom][
                        laser
                    ]["offset"]
                )
                if offset is not None:
                    remote_focus_offset += offset    

                # Calculate the Waveforms
                if sensor_mode == "Light-Sheet" and (
                    readout_direction == "Bidirectional"
                    or readout_direction == "Rev. Bidirectional"
                ):
                    sweep_time=self.sweep_time
                    amplitude = amplitude
                    offset = remote_focus_offset
                
                    self.triangle(sweep_time, amplitude, offset)

                else:
                    self.ramp(exposure_time, amplitude, remote_focus_offset)
    
    def triangle(
        self,
        sweep_time=0.24,
        amplitude=1,
        offset=0,
    ):
        """Sends the tiger controller commands to initiate the triangle wave.
        
        The waveform starts at the offset and immediately rises linearly to 2x amplitude 
        (amplitude here refers to 1/2 peak-to-peak) and immediately falls linearly back
        to the offset. The waveform is a periodic waveform with no delay periods between
        cycles.

        Parameters
        ----------
        sweep_time : Float
            Unit - Seconds
        amplitude : Float
            Unit - Volts
        offset : Float
            Unit - Volts
        """

        # Converts sweep_time to ms and amplitude and offset to mV
        period = int(round(sweep_time * 1000)) 
        amplitude *= 1000
        offset *= 1000

        # Triangle waveform
        self.remote_focus.SA_waveform(self.axis, 1, amplitude, offset, period)
        # Waveform is free running after it is triggered 
        self.remote_focus.SAM(self.axis, 4)

    def ramp(
        self,
        exposure_time=0.2,
        amplitude=1,
        offset=0.5,
    ):
        """Sends the tiger controller commands to initiate the ramp wave.

        The waveform starts at offset and immediately rises linearly to 2x amplitude 
        (amplitude here refers to 1/2 peak-to-peak) and then immediately drops back 
        down to the offset voltage during the fall period. 

        There is a delay period after each cycle that comes from the PLC, which is 
        not included in this function. 

        Parameters
        ----------
        exposure_time : Float
            Unit - Seconds
        sweep_time : Float
            Unit - Seconds
        remote_focus_delay : Float
            Unit - seconds
        camera_delay : Float
            Unit - seconds
        fall : Float
            Unit - seconds
        amplitude : Float
            Unit - Volts
        offset : Float
            Unit - Volts
        """

        # Converts exposure_time to ms and amplitude and offset to mV
        amplitude *= 1000
        offset *= 1000
        exposure_time = int(round(exposure_time * 1000))

        # Ramp waveform that is triggered on TTL inputs
        self.remote_focus.SA_waveform(self.axis, 128, amplitude, offset, exposure_time)
        # The waveform cycles once and waits for another TTL inputs
        self.remote_focus.SAM(self.axis, 2)
    
    def move(self, exposure_times, sweep_times, offset=None):
        """Move the remote focus.

        This method moves the remote focus.

        Parameters
        ----------
        exposure_times : dict
            Dictionary of exposure times for each selected channel
        sweep_times : dict
            Dictionary of sweep times for each selected channel
        offset : float
            The offset of the signal in volts.
        """
        
        self.adjust(exposure_times, sweep_times, offset)
    
    def turn_off(self): 
        """Stops the remote focus waveform"""
        self.remote_focus.SAM(self.axis, 0)

    def close(self):
        """Close the ASI remote_focus serial port.

        Stops the remote focus waveform and then closes the port.
        """
        if self.remote_focus.is_open():
            self.turn_off()
            logger.debug("ASI Remote Focus - Closing Device.")
            self.remote_focus.disconnect_from_serial()

    def __del__(self):
        """Destructor for the ASIRemoteFocus class."""
        self.close()

        



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Imports
import tkinter as tk
from tkinter import ttk
import logging
from pathlib import Path
from typing import Iterable, Optional

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.hover import HoverTkButton
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput
from navigate.view.custom_widgets.validation import ValidatedSpinbox
from navigate.view.custom_widgets.validation import ValidatedEntry
from navigate.view.custom_widgets.common import uniform_grid
import navigate

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class StageControlNotebook(ttk.Notebook):
    """Notebook for stage control tab."""

    def __init__(
        self,
        frame: "navigate.view.main_window_content."
        "settings_notebook.SettingsNotebook",
        *args: Iterable,
        **kwargs: dict,
    ) -> None:
        """Initialize the stage control notebook.

        Parameters
        ----------
        frame : SettingsNotebook
            The frame to put notebook into.
        *args : Iterable
            Arguments for ttk.Notebook
        **kwargs : Iterable
            Keyword arguments for ttk.Notebook
        """

        super().__init__()

        #: StageControlTab: Stage control tab.
        self.stage_control_tab = StageControlTab(self)

        # Adding tabs to notebook
        self.add(self.stage_control_tab, text="Stage Control", sticky=tk.NSEW)


class StageControlTab(tk.Frame):
    """Stage Control Tab for stage control notebook."""

    def __init__(
        self, note3: StageControlNotebook, *args: Iterable, **kwargs: dict
    ) -> None:
        """Initialize the stage control tab.

        Parameters
        ----------
        note3 : StageControlNotebook
            Stage control notebook.
        *args : Iterable
            Arguments for tk.Frame
        **kwargs : dict
            Keyword arguments for tk.Frame
        """
        tk.Frame.__init__(self, note3, *args, **kwargs)

        #: int: Index of the stage control tab.
        self.index = 2

        #: tk.PhotoImage: Image for the up button.
        self.up_1x_image = None

        #: tk.PhotoImage: Image for the 5x up button.
        self.up_5x_image = None

        #: tk.PhotoImage: Image for the down button.
        self.down_1x_image = None

        #: tk.PhotoImage: Image for the 5x down button.
        self.down_5x_image = None

        #: tk.PhotoImage: Image for the left button.
        self.left_1x_image = None

        #: tk.PhotoImage: Image for the 5x left button.
        self.left_5x_image = None

        #: tk.PhotoImage: Image for the right button.
        self.right_1x_image = None

        #: tk.PhotoImage: Image for the 5x right button.
        self.right_5x_image = None

        #: tk.PhotoImage: Image for the disabled up button.
        self.d_up_1x_image = None

        #: tk.PhotoImage: Image for the disabled 5x button.
        self.d_up_5x_image = None

        #: tk.PhotoImage: Image for the disabled down button.
        self.d_down_1x_image = None

        #: tk.PhotoImage: Image for the disabled 5x down button.
        self.d_down_5x_image = None

        #: tk.PhotoImage: Image for the disabled left button.
        self.d_left_1x_image = None

        #: tk.PhotoImage: Image for the disabled 5x left button.
        self.d_left_5x_image = None

        #: tk.PhotoImage: Image for the disabled right button.
        self.d_right_1x_image = None

        #: tk.PhotoImage: Image for the disabled 5x right button.
        self.d_right_5x_image = None

        self.load_images()

        #: PositionFrame: Position frame.
        self.position_frame = PositionFrame(self)
        self.position_frame.grid(row=0, column=0, rowspan=1, sticky=tk.NSEW, padx=3, pady=3)

        #: StackShortcuts: Stack shortcuts.
        self.stack_shortcuts = StackShortcuts(self)
        self.stack_shortcuts.grid(row=1, column=0, rowspan=1, sticky=tk.NSEW)

        #: XYFrame: XY frame.
        self.xy_frame = XYFrame(self)
        self.xy_frame.grid(row=0, column=1, rowspan=2, sticky=tk.NSEW, padx=3, pady=3)

        #: OtherAxisFrame: Z frame.
        self.z_frame = OtherAxisFrame(stage_control_tab=self, name="Z")
        self.z_frame.grid(row=0, column=2, rowspan=2, sticky=tk.NSEW, padx=3, pady=3)

        #: OtherAxisFrame: Theta frame.
        self.theta_frame = OtherAxisFrame(stage_control_tab=self, name="Theta")
        self.theta_frame.grid(row=2, column=2, rowspan=2, sticky=tk.NSEW, padx=3, pady=3)

        # OtherAxisFrame: Focus frame.
        self.f_frame = OtherAxisFrame(stage_control_tab=self, name="Focus")
        self.f_frame.grid(row=2, column=0, rowspan=2, sticky=tk.NSEW, padx=3, pady=3)

        #: StopFrame: Stop frame.
        self.stop_frame = StopFrame(
            stage_control_tab=self, name="Stage Movement Interrupt"
        )
        self.stop_frame.grid(row=2, column=1, rowspan=2, sticky=tk.NSEW, padx=3, pady=3)

        uniform_grid(self)

        self.default_axes = ["x", "y", "z", "theta", "f"]

    def load_images(self) -> None:
        """Load images for the stage control tab."""

        # Path to arrows
        image_directory = Path(__file__).resolve().parent
        image_names = [
            "up_1x_image",
            "up_5x_image",
            "down_1x_image",
            "down_5x_image",
            "left_1x_image",
            "left_5x_image",
            "right_1x_image",
            "right_5x_image",
            "d_up_1x_image",
            "d_up_5x_image",
            "d_down_1x_image",
            "d_down_5x_image",
            "d_left_1x_image",
            "d_left_5x_image",
            "d_right_1x_image",
            "d_right_5x_image",
        ]

        for name in image_names:
            setattr(
                self,
                f"{name}",
                tk.PhotoImage(
                    file=image_directory.joinpath("images", f"{name}.png")
                ).subsample(2, 2),
            )

    def get_widgets(self) -> dict:
        """Get all widgets in the stage control tab.

        Returns
        -------
        widgets: dict
            Dictionary of widgets
        """
        temp = {**self.position_frame.get_widgets()}
        for axis in self.default_axes:
            if axis == "x" or axis == "y":
                axis = "xy"
            temp[axis + "_step"] = getattr(self, axis + "_frame").get_widget()
        return temp

    def get_variables(self) -> dict:
        """Get all variables in the stage control tab.

        Returns
        -------
        variables: dict
            Dictionary of variables
        """
        temp = self.get_widgets()
        return {k: temp[k].get_variable() for k in temp}

    def get_buttons(self) -> dict:
        """Get all buttons in the stage control tab.

        Returns
        -------
        buttons: dict
            Dictionary of buttons
        """
        result = {**self.xy_frame.get_buttons()}
        for axis in self.default_axes:
            if axis == "x" or axis == "y":
                continue
            temp = getattr(self, axis + "_frame").get_buttons()
            result.update({k + "_" + axis + "_btn": temp[k] for k in temp})
        result.update(self.stop_frame.get_buttons())
        return result

    def toggle_button_states(
        self, joystick_is_on: bool = False, joystick_axes: Optional[list] = None
    ):
        """Enables/disables buttons and entries in the stage control tab,
        according to joystick axes.

        Parameters
        ----------
        joystick_is_on: bool
            'True' indicates that joystick mode is on
            'False' indicates that joystick mode is off
        joystick_axes: Optional[list]
            A list containing the axes controlled by the joystick, if any
        """
        if joystick_axes is None:
            joystick_axes = []

        self.xy_frame.toggle_button_states(joystick_is_on, joystick_axes)
        for axis in self.default_axes:
            if axis == "x" or axis == "y":
                continue
            getattr(self, axis + "_frame").toggle_button_states(
                joystick_is_on, joystick_axes
            )
        self.position_frame.toggle_entry_states(joystick_is_on, joystick_axes)
        self.stop_frame.toggle_button_states(joystick_is_on, joystick_axes)

    def force_enable_all_axes(self) -> None:
        """Enable all buttons and entries in the stage control tab."""
        self.xy_frame.toggle_button_states(False, ["x", "y"])
        for axis in self.default_axes:
            if axis == "x" or axis == "y":
                continue
            getattr(self, axis + "_frame").toggle_button_states(False, [axis])
        self.position_frame.toggle_entry_states(False, self.default_axes)
        self.stop_frame.toggle_button_states(False, [])

    def add_additional_stage(self, stage_name: str) -> None:
        """Add an additional stage to the stage control tab.

        Parameters
        ----------
        stage_name: str
            Name of the additional stage
        """
        self.default_axes.append(stage_name)
        additional_stage = OtherAxisFrame(self, stage_name.upper())
        additional_stage.grid(
            row=(len(self.default_axes) % 2) * 2,
            column=len(self.default_axes) // 2 + 1,
            sticky=tk.NSEW,
            rowspan=2,
            padx=3,
            pady=3,
        )
        setattr(self, f"{stage_name}_frame", additional_stage)

        self.position_frame.add_position_entry(stage_name, stage_name.upper())

        uniform_grid(self)


class OtherAxisFrame(ttk.Labelframe):
    """Frame for the other axis movement buttons.

    This frame is used for the z, theta, and focus axis movement buttons.
    """

    def __init__(
        self,
        stage_control_tab: StageControlTab,
        name: str,
        *args: Iterable,
        **kwargs: dict,
    ) -> None:
        """Initialize the other axis frame.

        Parameters
        ----------
        stage_control_tab : StageControlTab
            The stage control tab that the other axis frame is in
        name : str
            The name of the axis that the frame is for
        *args : Iterable
            Positional arguments for the ttk.Labelframe
        **kwargs : Iterable
            Keyword arguments for the ttk.Labelframe
        """

        ttk.Labelframe.__init__(
            self,
            stage_control_tab,
            text=name + " Movement",
            labelanchor="n",
            *args,
            **kwargs,
        )

        #: str: Name of the axis.
        self.name = name

        #: StageControlTab: Stage control tab.
        self.stage_control_tab = stage_control_tab

        #: HoverTkButton: Up button.
        self.up_btn = HoverTkButton(
            self, image=self.stage_control_tab.d_up_1x_image, borderwidth=0
        )

        #: HoverTkButton: 5x Up button.
        self.large_up_btn = HoverTkButton(
            self, image=self.stage_control_tab.up_5x_image, borderwidth=0
        )

        #: HoverTkButton: Down button.
        self.down_btn = HoverTkButton(
            self, image=self.stage_control_tab.down_1x_image, borderwidth=0
        )

        #: HoverTkButton: 5x Down button.
        self.large_down_btn = HoverTkButton(
            self, image=self.stage_control_tab.down_5x_image, borderwidth=0
        )

        if self.name.lower() == "theta":
            text = "Step Size (" + "\N{DEGREE SIGN}" + ")"
        else:
            text = "Step Size (" + "\N{GREEK SMALL LETTER MU}" + "m)"

        # #: LabelInput: Increment spinbox.
        self.increment_box = LabelInput(
            parent=self,
            input_class=ValidatedSpinbox,
            input_var=tk.DoubleVar(),
            input_args={"width": 5},
            label=text,
            label_pos="top",
            label_args={"font": tk.font.Font(size=10)},
        )

        # # Adding space between buttons
        space_1 = ttk.Label(self, borderwidth=0)
        space_2 = ttk.Label(self, borderwidth=0)

        # Griding out buttons
        self.large_up_btn.grid(
            row=(row := 0), column=1, rowspan=1, columnspan=1, padx=2, pady=2
        )
        self.up_btn.grid(
            row=(row := row + 1), column=1, rowspan=1, columnspan=1, pady=(2, 0)
        )
        space_1.grid(
            row=(row := row + 1), column=1, rowspan=1, columnspan=1, padx=2, pady=0
        )
        self.increment_box.grid(
            row=(row := row + 1), column=1, rowspan=1, columnspan=1, padx=2, pady=0
        )
        space_2.grid(
            row=(row := row + 1), column=1, rowspan=1, columnspan=1, padx=2, pady=0
        )
        self.down_btn.grid(
            row=(row := row + 1), column=1, rowspan=1, columnspan=1, pady=(0, 2)
        )
        self.large_down_btn.grid(
            row=(row + 1), column=1, rowspan=1, columnspan=1, padx=2, pady=2
        )

        uniform_grid(self)


    def get_widget(self) -> LabelInput:
        """Returns the frame widget

        Returns
        -------
        increment_box: LabelInput
            The increment widget
        """
        return self.increment_box

    def get_buttons(self) -> dict:
        """Returns the buttons in the frame

        Returns
        -------
        buttons: dict
            A dictionary of the buttons
        """
        return {
            "up": self.up_btn,
            "down": self.down_btn,
            "large_up": self.large_up_btn,
            "large_down": self.large_down_btn,
        }

    def toggle_button_states(
        self, joystick_is_on: bool = False, joystick_axes: Optional[list] = None
    ):
        """Switches the images used as buttons between two states

        Parameters
        ----------
        joystick_is_on : bool
            'True' indicates that joystick mode is on
            'False' indicates that joystick mode is off
        joystick_axes : Optional[list]
            A list containing the axes controlled by the joystick, if any
        """

        normal_images = [
            self.stage_control_tab.up_1x_image,
            self.stage_control_tab.down_1x_image,
            self.stage_control_tab.down_5x_image,
            self.stage_control_tab.up_5x_image,
        ]

        disabled_images = [
            self.stage_control_tab.d_up_1x_image,
            self.stage_control_tab.d_down_1x_image,
            self.stage_control_tab.d_down_5x_image,
            self.stage_control_tab.d_up_5x_image,
        ]

        buttons = [
            self.up_btn,
            self.down_btn,
            self.large_down_btn,
            self.large_up_btn,
        ]

        if joystick_axes is None:
            joystick_axes = []

        # Default Button State
        button_state = "normal"
        image_list = normal_images

        if joystick_is_on:
            if (self.name.lower() in joystick_axes) or (
                self.name.lower() == "focus" and "f" in joystick_axes
            ):
                button_state = "disabled"
                image_list = disabled_images

        for k in range(len(buttons)):
            buttons[k]["state"] = button_state
            buttons[k].config(image=image_list[k])


class PositionFrame(ttk.Labelframe):
    """Frame for the stage position entries.

    This frame is used for the x, y, z, theta, and focus position entries.
    """

    def __init__(
        self, stage_control_tab: StageControlTab, *args: Iterable, **kwargs: dict
    ) -> None:
        """Initialize the position frame.

        Parameters
        ----------
        stage_control_tab : StageControlTab
            The stage control tab that the position frame is in
        *args : Iterable
            Positional arguments for the ttk.Labelframe
        **kwargs : Iterable
            Keyword arguments for the ttk.Labelframe
        """

        ttk.Labelframe.__init__(
            self, stage_control_tab, text="Stage Positions", *args, **kwargs
        )

        #: dict: Dictionary of the label input widgets for the position entries.
        self.inputs = {}

        #: list: List of frames for the position entries.
        entry_names = ["x", "y", "z", "theta", "f"]

        #: list: List of labels for the position entries.
        entry_labels = ["X", "Y", "Z", "\N{Greek Capital Theta Symbol}", "F"]

        #: ttk.Style: Style for the position entries.
        self.position_style = ttk.Style()
        self.position_style.configure("Position.TEntry", fieldbackground="white")

        #: list: List of frames for the position entries.
        for i in range(len(entry_names)):
            self.add_position_entry(entry_names[i], entry_labels[i])

        uniform_grid(self)


    def get_widgets(self) -> dict:
        """Get all widgets in the position frame

        Returns
        -------
        inputs: dict
            A dictionary of the label input widgets for the position entries
        """

        return self.inputs

    def get_variables(self) -> dict:
        """Get all variables in the position frame

        Returns
        -------
        variables: dict
            A dictionary of the variables for the position entries
        """

        variables = {}
        for name in self.inputs:
            variables[name] = self.inputs[name].get_variable()
        return variables

    def toggle_entry_states(
        self, joystick_is_on: bool = False, joystick_axes: Optional[list] = None
    ) -> None:
        """Switches the images used as buttons between two states

        Parameters
        ----------
        joystick_is_on : bool
            'True' indicates that joystick mode is on
            'False' indicates that joystick mode is off
        joystick_axes : Optional[list]
            A list containing the axes controlled by the joystick, if any
        """

        if joystick_axes is None:
            joystick_axes = []

        frame_back_counter = 0
        if joystick_is_on:
            entry_state = "disabled"
            frame_back_color = "#ee868a"
        else:
            entry_state = "normal"
            frame_back_color = "#f0f0f0"
        self.position_style.configure("Position.TEntry", fieldbackground=frame_back_color)

        for variable in self.get_variables():
            if variable in joystick_axes:
                try:
                    self.inputs[f"{variable}"].widget.config(style="Position.TEntry")
                    self.inputs[f"{variable}"].widget["state"] = entry_state

                except KeyError:
                    pass
            frame_back_counter += 1

    def add_position_entry(self, entry_name: str, entry_label: str) -> None:
        """Add a position entry to the position frame.

        Parameters
        ----------
        entry_name : str
            Name of the entry
        entry_label : str
            Label of the entry
        """

        self.inputs[entry_name] = LabelInput(
            parent=self,
            label=entry_label,
            input_class=ValidatedEntry,
            input_var=tk.StringVar(),
            input_args={
                "required": True,
                "precision": 0.1,
                "takefocus": False,
            },
        )
        self.inputs[entry_name].grid(row=len(self.inputs), column=0, sticky=tk.EW)

        uniform_grid(self)


class StackShortcuts(ttk.LabelFrame):
    def __init__(self, position_frame: PositionFrame, *args: Iterable, **kwargs: dict) -> None:
        """Initialize the stack shortcuts frame.

        Parameters
        ----------
        position_frame : PositionFrame
            The position frame that the stack shortcuts frame is in
        *args : Iterable
            Positional arguments for the ttk.Labelframe
        **kwargs : Iterable
            Keyword arguments for the ttk.Labelframe
        """
        ttk.Labelframe.__init__(
            self, position_frame, text="Z-Stack Start/Stop", *args, **kwargs
        )

        # Add two buttons
        self.set_start_button = HoverTkButton(self, text="Set Start Pos/Foc")
        self.set_start_button.grid(row=0, column=0, sticky="ew")

        self.set_end_button = HoverTkButton(self, text="Set End Pos/Foc")
        self.set_end_button.grid(row=1, column=0, sticky="ew")

        uniform_grid(self)


class XYFrame(ttk.Labelframe):
    """Frame for the x and y movement buttons.

    This frame is used for the up, down, left, right, and increment buttons.
    """

    def __init__(
        self, stage_control_tab: StageControlTab, *args: Iterable, **kwargs: dict
    ) -> None:
        """Initialize the XY frame.

        Parameters
        ----------
        stage_control_tab : StageControlTab
            The stage control tab that the XY frame is in
        *args : Iterable
            Positional arguments for the ttk.Labelframe
        **kwargs : dict
            Keyword arguments for the ttk.Labelframe
        """

        ttk.Labelframe.__init__(
            self,
            stage_control_tab,
            text="X Y Movement",
            labelanchor="n",
            *args,
            **kwargs,
        )

        #: StageControlTab: Stage control tab.
        self.stage_control_tab = stage_control_tab

        #: HoverTkButton: Up button.
        self.up_y_btn = HoverTkButton(
            self, image=self.stage_control_tab.up_1x_image, borderwidth=0
        )

        #: HoverTkButton: Up button.
        self.large_up_y_btn = HoverTkButton(
            self, image=self.stage_control_tab.up_5x_image, borderwidth=0
        )

        #: HoverTkButton: Down button.
        self.down_y_btn = HoverTkButton(
            self, image=self.stage_control_tab.down_1x_image, borderwidth=0
        )

        #: HoverTkButton: Down button.
        self.large_down_y_btn = HoverTkButton(
            self, image=self.stage_control_tab.down_5x_image, borderwidth=0
        )

        #: HoverTkButton: Right button.
        self.up_x_btn = HoverTkButton(
            self, image=self.stage_control_tab.right_1x_image, borderwidth=0
        )

        #: HoverTkButton: Right 5x button.
        self.large_up_x_btn = HoverTkButton(
            self, image=self.stage_control_tab.right_5x_image, borderwidth=0
        )

        #: HoverTkButton: Left button.
        self.down_x_btn = HoverTkButton(
            self, image=self.stage_control_tab.left_1x_image, borderwidth=0
        )

        #: HoverTkButton: Left 5x button.
        self.large_down_x_btn = HoverTkButton(
            self, image=self.stage_control_tab.left_5x_image, borderwidth=0
        )

        #: LabelInput: Increment spinbox.
        self.increment_box = LabelInput(
            parent=self,
            input_class=ValidatedSpinbox,
            input_var=tk.DoubleVar(),
            input_args={"width": 5},
            label="Step Size (\N{GREEK SMALL LETTER MU}m)",
            label_pos="top",
            label_args={"font": tk.font.Font(size=10)},
        )

        #: dict: Dictionary of the buttons for the x and y movement buttons.
        self.button_axes_dict = {
            "x": [
                self.up_x_btn,
                self.down_x_btn,
                self.large_up_x_btn,
                self.large_down_x_btn,
            ],
            "y": [
                self.up_y_btn,
                self.down_y_btn,
                self.large_up_y_btn,
                self.large_down_y_btn,
            ],
        }

        # Up
        self.large_up_y_btn.grid(
            row=0, column=4, rowspan=2, columnspan=2, padx=2, pady=2
        )

        self.up_y_btn.grid(row=2, column=4, rowspan=2, columnspan=2, padx=2, pady=2)

        # Increment box.
        self.increment_box.grid(
            row=4, column=4, rowspan=3, columnspan=2, padx=2, pady=2
        )

        # Down
        self.down_y_btn.grid(row=7, column=4, rowspan=2, columnspan=2, padx=2, pady=2)

        self.large_down_y_btn.grid(
            row=9, column=4, rowspan=2, columnspan=2, padx=2, pady=2
        )

        # Left
        self.large_down_x_btn.grid(
            row=5, column=0, rowspan=2, columnspan=2, padx=2, pady=2
        )

        self.down_x_btn.grid(
            row=5,
            column=2,
            rowspan=2,
            columnspan=2,
            padx=2,
            pady=2,
        )

        # Right
        self.up_x_btn.grid(row=5, column=6, rowspan=2, columnspan=2, padx=2, pady=2)
        self.large_up_x_btn.grid(
            row=5, column=8, rowspan=2, columnspan=2, padx=2, pady=2
        )

        uniform_grid(self)

        # Increment spinbox
        self.increment_box.widget.set_precision(-1)

    def get_widget(self) -> LabelInput:
        """Returns the frame widget

        Returns
        -------
        increment_box: LabelInput
            The frame widget
        """

        return self.increment_box

    def get_buttons(self) -> dict:
        """Returns the buttons in the frame

        Returns
        -------
        buttons: dict
            A dictionary of the buttons
        """

        names = [
            "up_x_btn",
            "down_x_btn",
            "up_y_btn",
            "down_y_btn",
            "large_up_x_btn",
            "large_down_x_btn",
            "large_up_y_btn",
            "large_down_y_btn",
        ]
        return {k: getattr(self, k) for k in names}

    def toggle_button_states(
        self, joystick_is_on: bool = False, joystick_axes: Optional[list] = None
    ) -> None:
        """Switches the images used as buttons between two states

        joystick_is_on : bool
            False if buttons are normal, True if buttons are disabled
        joystick_axes : Optional[list]
            A list of the joystick axes
        """

        if joystick_axes is None:
            joystick_axes = []

            # Default Button State
        button_state = "normal"
        image_list = {
            "x": [
                self.stage_control_tab.right_1x_image,
                self.stage_control_tab.left_1x_image,
                self.stage_control_tab.right_5x_image,
                self.stage_control_tab.left_5x_image,
            ],
            "y": [
                self.stage_control_tab.up_1x_image,
                self.stage_control_tab.down_1x_image,
                self.stage_control_tab.up_5x_image,
                self.stage_control_tab.down_5x_image,
            ],
        }

        if joystick_is_on:
            if "x" in joystick_axes:
                button_state = "disabled"
                image_list["x"] = [
                    self.stage_control_tab.d_right_1x_image,
                    self.stage_control_tab.d_left_1x_image,
                    self.stage_control_tab.d_right_5x_image,
                    self.stage_control_tab.d_left_5x_image,
                ]
            if "y" in joystick_axes:
                button_state = "disabled"
                image_list["y"] = [
                    self.stage_control_tab.d_up_1x_image,
                    self.stage_control_tab.d_down_1x_image,
                    self.stage_control_tab.d_up_5x_image,
                    self.stage_control_tab.d_down_5x_image,
                ]

        for k, button in enumerate(self.button_axes_dict["x"]):
            button["state"] = button_state
            button.config(image=image_list["x"][k])

        for k, button in enumerate(self.button_axes_dict["y"]):
            button["state"] = button_state
            button.config(image=image_list["y"][k])


class StopFrame(ttk.Labelframe):
    """Frame for the stop button."""

    def __init__(
        self,
        stage_control_tab: StageControlTab,
        name: str,
        *args: Iterable,
        **kwargs: dict,
    ) -> None:
        """Initialize the stop frame.

        Parameters
        ----------
        stage_control_tab : StageControlTab
            Stage control tab.
        name : str
            Name of the frame.
        *args
            Arguments for ttk.Frame
        **kwargs
            Keyword arguments for ttk.Frame
        """

        # Init Frame
        ttk.Labelframe.__init__(
            self, stage_control_tab, text=name, labelanchor="n", *args, **kwargs
        )

        #: tk.Button: Stop button.
        self.stop_btn = tk.Button(
            self, bg="red", fg="white", text="STOP", width=20, height=6
        )

        #: HoverTkButton: Joystick button.
        self.joystick_btn = HoverTkButton(
            self, bg="white", fg="black", text="Enable Joystick", width=15, height=2
        )

        # Griding out buttons
        self.stop_btn.grid(row=0, column=0, rowspan=2, pady=2)
        self.joystick_btn.grid(row=2, column=0, rowspan=2, pady=2)

        uniform_grid(self)

    def get_buttons(self) -> dict:
        """Returns the buttons in the frame

        Returns
        -------
        buttons: dict
            A dictionary of the buttons
        """
        return {"stop": self.stop_btn, "joystick": self.joystick_btn}

    def toggle_button_states(
        self, joystick_is_on: bool = False, joystick_axes: Optional[list] = None
    ):
        """Switches the images used as buttons between two states

        Parameters
        ----------
        joystick_is_on : bool
            'True' indicates that joystick mode is on
            'False' indicates that joystick mode is off
        joystick_axes : Optional[list]
            A list containing the axes controlled by the joystick, if any
        """
        if joystick_is_on:
            self.joystick_btn.config(text="Disable Joystick")
        else:
            self.joystick_btn.config(text="Enable Joystick")



""" Hardware devices. """

# Standard library imports

# Third party imports

# Local imports
from .daq.synthetic import SyntheticDAQ  # noqa
from .camera.synthetic import SyntheticCamera  # noqa
from .filter_wheel.synthetic import SyntheticFilterWheel  # noqa
from .galvo.synthetic import SyntheticGalvo  # noqa
from .remote_focus.synthetic import SyntheticRemoteFocus  # noqa
from .shutter.synthetic import SyntheticShutter  # noqa
from .stage.synthetic import SyntheticStage  # noqa
from .zoom.synthetic import SyntheticZoom  # noqa
from .laser.synthetic import SyntheticLaser  # noqa
from .mirror.synthetic import SyntheticMirror  # noqa



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard Library Imports
import logging
from tkinter import ttk
import tkinter as tk
from tkinter.scrolledtext import ScrolledText
import platform

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.popup import PopUp
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput
from navigate.view.custom_widgets.validation import ValidatedCombobox, ValidatedSpinbox
from navigate.model.data_sources import FILE_TYPES
from navigate.view.custom_widgets.common import CommonMethods

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


SOLVENTS = ("BABB", "Water", "CUBIC", "CLARITY", "uDISCO", "eFLASH")


class AcquirePopUp(CommonMethods):
    """Class creates the popup that is generated when the Acquire button is pressed and
    Save File checkbox is selected."""

    def __init__(self, root: tk.Tk) -> None:
        """Initialize the AcquirePopUp class

        Parameters
        ----------
        root : tk.Tk
            The root window
        """
        #: tk.Tk: The root window
        self.tk = root

        #: int: Width of the first column
        self.column1_width = 20

        #: int: Width of the second column
        self.column2_width = 40

        #: PopUp: The popup window
        if platform.system() == "Windows":
            self.global_width = 450
            self.global_height = 710
            self.popup = PopUp(
                root,
                name="File Saving Dialog",
                size=f"{self.global_width}x{self.global_height}+320+180",
                transient=True,
            )
        else:
            self.global_width = 580
            self.global_height = 730
            self.popup = PopUp(
                root,
                name="File Saving Dialog",
                size=f"{self.global_width}x{self.global_height}+320+180",
                transient=True,
            )

        #: dict: Button dictionary.
        self.buttons = {}

        #: dict: Input dictionary.
        self.inputs = {}

        content_frame = self.popup.get_frame()
        content_frame.columnconfigure(index=0, weight=1)
        content_frame.rowconfigure(index=0, weight=1)

        path_entries = ttk.Frame(content_frame, padding=(5, 5, 5, 5))
        tab_frame = ttk.Frame(content_frame, padding=(5, 5, 5, 5))
        button_frame = ttk.Frame(content_frame, padding=(5, 5, 5, 5))
        separator1 = ttk.Separator(content_frame, orient="horizontal")
        separator2 = ttk.Separator(content_frame, orient="horizontal")

        path_entries.grid(row=0, column=0, sticky=tk.NSEW, padx=0, pady=3)
        path_entries.grid_columnconfigure(index=0, weight=1)
        path_entries.grid_rowconfigure(index=1, weight=1)

        separator1.grid(row=1, column=0, sticky=tk.NSEW, padx=0, pady=3)

        tab_frame.grid(row=2, column=0, sticky=tk.NSEW, padx=0, pady=3)
        tab_frame.grid_columnconfigure(index=0, weight=1)
        tab_frame.grid_rowconfigure(index=1, weight=1)

        separator2.grid(row=3, column=0, sticky=tk.NSEW, padx=0, pady=3)

        button_frame.grid(row=4, column=0, sticky=tk.NSEW, padx=0, pady=3)
        button_frame.grid_columnconfigure(index=0, weight=1)
        button_frame.grid_rowconfigure(index=1, weight=1)

        #: ButtonFrame: ButtonFrame object
        self.button_frame = ButtonFrame(parent=self, frame=button_frame)

        #: EntryFrame: EntryFrame object
        self.path_frame = EntryFrame(parent=self, frame=path_entries)

        #: TabFrame: TabFrame object
        self.tab_frame = TabFrame(parent=self, frame=tab_frame)


class ButtonFrame:
    def __init__(self, parent: AcquirePopUp, frame: ttk.Frame) -> None:
        """Initialize the ButtonFrame

        Parameters
        ----------
        parent : AcquirePopUp
            The AcquirePopup window.
        frame : ttk.Frame
            The AcquirePopup Window.
        """

        width = int((parent.column1_width + parent.column2_width - 10) / 2)
        parent.buttons["Cancel"] = ttk.Button(
            frame, text="Cancel Acquisition", width=width
        )
        parent.buttons["Done"] = ttk.Button(frame, text="Acquire Data", width=width)
        parent.buttons["Cancel"].grid(row=0, column=0, padx=5, sticky=tk.NSEW)
        parent.buttons["Done"].grid(row=0, column=1, padx=5, sticky=tk.NSEW)


class EntryFrame:
    def __init__(self, parent: AcquirePopUp, frame: ttk.Frame) -> None:
        """Initialize the EntryFrame

        Parameters
        ----------
        parent : AcquirePopUp
            The AcquirePopup window.
        frame : ttk.Frame
            The EntryFrame Window.
        """
        row_index = 0
        text = "Please Fill Out the Fields Below"

        #: ttk.Label: Label for the entries
        label = ttk.Label(frame, text=text)
        label.grid(row=0, column=0, columnspan=2, sticky=tk.NSEW, pady=5, padx=0)

        # Creating Entry Widgets
        entry_names = [
            "root_directory",
            "user",
            "tissue",
            "celltype",
            "label",
            "prefix",
            "solvent",
            "file_type",
        ]

        entry_labels = [
            "Root Directory",
            "User",
            "Tissue Type",
            "Cell Type",
            "Label",
            "Prefix",
            "Solvent",
            "File Type",
        ]

        # Loop for each entry and label
        row_index += 1
        for i in range(len(entry_names)):
            if entry_names[i] == "file_type":
                parent.inputs[entry_names[i]] = LabelInput(
                    parent=frame,
                    label=entry_labels[i],
                    input_class=ValidatedCombobox,
                    input_var=tk.StringVar(),
                )
                parent.inputs[entry_names[i]].widget.state(["!disabled", "readonly"])
                parent.inputs[entry_names[i]].set_values(tuple(FILE_TYPES))
                parent.inputs[entry_names[i]].set("TIFF")

            elif entry_names[i] == "solvent":
                parent.inputs[entry_names[i]] = LabelInput(
                    parent=frame,
                    label=entry_labels[i],
                    input_class=ValidatedCombobox,
                    input_var=tk.StringVar(),
                )
                parent.inputs[entry_names[i]].widget.state(["!disabled", "readonly"])
                parent.inputs[entry_names[i]].set_values(SOLVENTS)
                parent.inputs[entry_names[i]].set("BABB")

            else:
                parent.inputs[entry_names[i]] = LabelInput(
                    parent=frame,
                    label=entry_labels[i],
                    input_class=ttk.Entry,
                    input_var=tk.StringVar(),
                    input_args={"width": parent.column2_width},
                )

            # Widgets
            parent.inputs[entry_names[i]].grid(
                row=row_index,
                column=0,
                columnspan=1,
                sticky=tk.NSEW,
                padx=(0, 0),
                pady=(1, 1),
            )

            # Labels
            parent.inputs[entry_names[i]].label.grid(padx=(5, 5))
            parent.inputs[entry_names[i]].label.config(width=parent.column1_width)
            parent.inputs[entry_names[i]].widget.grid(padx=(0, 0), pady=(1, 1))
            row_index += 1


class TabFrame:
    def __init__(self, parent: AcquirePopUp, frame: ttk.Frame) -> None:
        """Initialize the TabFrame

        Parameters
        ----------
        parent : AcquirePopUp
            The AcquirePopup window.
        frame : ttk.Frame
            The TabFrame Window.
        """
        notebook = ttk.Notebook(frame, padding=(5, 2, 5, 2))
        notebook.grid(row=0, column=0, sticky=tk.NSEW)

        tab1 = tk.Frame(notebook)
        tab1.columnconfigure(index=0, weight=1)

        tab2 = tk.Frame(notebook)
        tab2.columnconfigure(index=0, weight=1)
        tab2.columnconfigure(index=1, weight=1)
        tab2.columnconfigure(index=2, weight=1)

        notebook.add(tab1, text="Misc. Notes")
        notebook.add(tab2, text="BDV Settings")

        row_index = 0

        text = "All notes are saved in to the header of the image file."

        notes_label = tk.Label(
            tab1,
            text=text,
            justify=tk.LEFT,
            width=parent.global_width - 30,
            wraplength=parent.global_width - 30,
        )

        notes_label.grid(
            row=row_index,
            column=0,
            columnspan=3,
            rowspan=2,
            sticky=tk.NSEW,
            pady=(5, 5),
        )

        row_index += 2
        separator1 = ttk.Separator(tab1, orient="horizontal")

        separator1.grid(
            row=row_index, column=0, columnspan=3, sticky=tk.NSEW, padx=0, pady=3
        )

        row_index += 1
        self.inputs = {
            "misc": ScrolledText(
                tab1,
                wrap=tk.WORD,
                height=20,
                width=parent.column2_width + parent.column2_width - 35,
            )
        }

        self.inputs["misc"].grid(row=row_index, column=0, columnspan=1, sticky=tk.NSEW)

        row_index = 0
        text = (
            "HDF5, N5, and Zarr files are saved with BDV metadata, "
            "enabling immediate visualization with BigDataViewer. "
            "All angles are in degrees."
        )

        bdv_label = tk.Label(
            tab2,
            text=text,
            justify=tk.LEFT,
            width=parent.global_width - 40,
            wraplength=parent.global_width - 40,
        )

        bdv_label.grid(
            row=row_index,
            column=0,
            columnspan=3,
            rowspan=2,
            sticky=tk.NSEW,
            pady=(5, 5),
        )

        row_index += 2
        self.inputs["shear_data"] = LabelInput(
            parent=tab2,
            label_pos="left",
            label="Shear",
            input_class=ttk.Checkbutton,
            input_var=tk.BooleanVar(),
            input_args={"onvalue": True, "offvalue": False},
        )
        self.inputs["shear_data"].grid(
            row=row_index,
            column=0,
            columnspan=1,
            sticky=tk.NSEW,
            padx=(5, 5),
            pady=(1, 1),
        )

        values = ["XZ", "YZ", "XY"]
        self.inputs["shear_dimension"] = LabelInput(
            parent=tab2,
            label_pos="top",
            label="Dimension",
            input_class=ttk.Combobox,
            input_var=tk.StringVar(),
            input_args={"values": values, "state": "readonly"},
        )

        self.inputs["shear_dimension"].grid(
            row=row_index, column=1, columnspan=1, sticky=tk.W, padx=(5, 5), pady=(1, 1)
        )

        self.inputs["shear_angle"] = LabelInput(
            parent=tab2,
            label_pos="top",
            label="Angle",
            input_class=ValidatedSpinbox,
            input_var=tk.StringVar(),
            input_args={
                "from_": 0,
                "to": 360,
                "increment": 1,
            },
        )
        self.inputs["shear_angle"].grid(
            row=row_index, column=2, columnspan=1, sticky=tk.W, padx=(5, 5), pady=(1, 1)
        )

        row_index += 1
        separator1 = ttk.Separator(tab2, orient="horizontal")

        separator1.grid(
            row=row_index, column=0, columnspan=3, sticky=tk.NSEW, padx=0, pady=3
        )

        row_index += 1
        self.inputs["rotate_data"] = LabelInput(
            parent=tab2,
            label_pos="left",
            label="Rotate",
            input_class=ttk.Checkbutton,
            input_var=tk.BooleanVar(),
            input_args={"onvalue": True, "offvalue": False},
        )

        self.inputs["rotate_data"].grid(
            row=row_index,
            column=0,
            columnspan=1,
            sticky=tk.NSEW,
            padx=(5, 5),
            pady=(1, 1),
        )

        # Insert a new frame here, and then add the widgets to that frame
        rotate_notebook = ttk.Notebook(tab2, padding=(5, 2, 5, 2))
        rotate_notebook.grid(row=row_index, column=1, columnspan=2, sticky=tk.NSEW)
        for i in range(3):
            rotate_notebook.columnconfigure(index=i, weight=1)

        self.inputs["rotate_angle_x"] = LabelInput(
            parent=rotate_notebook,
            label_pos="top",
            label="X Angle",
            input_class=ValidatedSpinbox,
            input_var=tk.StringVar(),
            input_args={
                "from_": 0,
                "to": 360,
                "increment": 1,
            },
        )

        self.inputs["rotate_angle_x"].grid(
            row=0, column=0, columnspan=1, sticky=tk.W, padx=(5, 5), pady=(1, 1)
        )

        self.inputs["rotate_angle_y"] = LabelInput(
            parent=rotate_notebook,
            label_pos="top",
            label="Y Angle",
            input_class=ValidatedSpinbox,
            input_var=tk.StringVar(),
            input_args={
                "from_": 0,
                "to": 360,
                "increment": 1,
            },
        )
        self.inputs["rotate_angle_y"].grid(
            row=0, column=1, columnspan=1, sticky=tk.W, padx=(5, 5), pady=(1, 1)
        )

        self.inputs["rotate_angle_z"] = LabelInput(
            parent=rotate_notebook,
            label_pos="top",
            label="Y Angle",
            input_class=ValidatedSpinbox,
            input_var=tk.StringVar(),
            input_args={
                "from_": 0,
                "to": 360,
                "increment": 1,
            },
        )
        self.inputs["rotate_angle_z"].grid(
            row=0, column=2, columnspan=1, sticky=tk.W, padx=(5, 5), pady=(1, 1)
        )

        row_index += 1
        separator1 = ttk.Separator(tab2, orient="horizontal")

        separator1.grid(
            row=row_index, column=0, columnspan=3, sticky=tk.NSEW, padx=0, pady=3
        )

        row_index += 1
        separator2 = ttk.Separator(tab2, orient="horizontal")
        separator2.grid(
            row=row_index, column=0, columnspan=3, sticky=tk.NSEW, padx=0, pady=3
        )

        row_index += 1
        self.inputs["down_sample_data"] = LabelInput(
            parent=tab2,
            label_pos="left",
            label="Downsample",
            input_class=ttk.Checkbutton,
            input_var=tk.BooleanVar(),
            input_args={"onvalue": True, "offvalue": False},
        )

        self.inputs["down_sample_data"].grid(
            row=row_index,
            column=0,
            columnspan=1,
            sticky=tk.NSEW,
            padx=(5, 5),
            pady=(1, 1),
        )

        values = ["1x", "2x", "4x", "8x", "16x", "32x", "64x", "128x"]
        self.inputs["lateral_down_sample"] = LabelInput(
            parent=tab2,
            label_pos="top",
            label="Lateral Downsample",
            input_class=ttk.Combobox,
            input_var=tk.StringVar(),
            input_args={"values": values, "state": "readonly"},
        )

        self.inputs["lateral_down_sample"].grid(
            row=row_index, column=1, columnspan=1, sticky=tk.W, padx=(5, 5), pady=(1, 1)
        )

        values = ["1x", "2x", "4x", "8x", "16x", "32x", "64x", "128x"]
        self.inputs["axial_down_sample"] = LabelInput(
            parent=tab2,
            label_pos="top",
            label="Axial Downsample",
            input_class=ttk.Combobox,
            input_var=tk.StringVar(),
            input_args={"values": values, "state": "readonly"},
        )

        self.inputs["axial_down_sample"].grid(
            row=row_index, column=2, columnspan=1, sticky=tk.W, padx=(5, 5), pady=(1, 1)
        )

        row_index += 1
        separator3 = ttk.Separator(tab2, orient="horizontal")
        separator3.grid(
            row=row_index, column=0, columnspan=3, sticky=tk.NSEW, padx=0, pady=3
        )

        row_index += 1
        text = (
            "Down-sampling is accompanied with additional computational overhead "
            "and slows data saving operations. Use with caution. The specified "
            "down-sampling is the maximum value, and intermediate values will "
            "automatically be calculated. "
        )

        bdv_label2 = tk.Label(
            tab2,
            text=text,
            justify=tk.LEFT,
            width=parent.global_width - 40,
            wraplength=parent.global_width - 40,
        )

        bdv_label2.grid(
            row=row_index,
            column=0,
            columnspan=3,
            rowspan=2,
            sticky=tk.NSEW,
            pady=(5, 5),
        )



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard Imports
import logging

# Third Party Imports

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)

""" File type specific data sources. """

FILE_TYPES = ["TIFF", "OME-TIFF", "H5", "N5", "OME-Zarr"]


def get_data_source(file_type: str):
    """Get the data source class for the given file type.

    Parameters
    ----------
    file_type : str
        File type to get the data source for.

    Returns
    -------
    DataSource
        Data source class for the given file type.
    """

    if (file_type == "TIFF") or (file_type == "OME-TIFF"):
        from .tiff_data_source import TiffDataSource

        return TiffDataSource

    elif (file_type == "H5") or file_type == "N5":
        from .bdv_data_source import BigDataViewerDataSource

        return BigDataViewerDataSource

    elif file_type == "OME-Zarr":
        from .zarr_data_source import OMEZarrDataSource

        return OMEZarrDataSource

    else:
        logger.error(f"Unknown file type {file_type}. Cannot open.")
        raise NotImplementedError(f"Unknown file type {file_type}. Cannot open.")



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

def slice_len(sl, n):
    """Calculate the length of the slice over an array of size n.

    Parameters
    ----------
    sl : slice
        The slice.
    n : int
        The size of the array.

    Returns
    -------
    int
        The length of the slice.
    """
    return len(range(n)[sl])

def key_len(keys):
    # Check lengths
    if isinstance(keys, slice) or isinstance(keys, int):
        length = 1
    else:
        length = len(keys)

    if length < 1:
        raise IndexError(
            "Too few indices."
        )

    return length

def ensure_iter(keys, pos, shape):
    """Ensure the output is iterable.

    Parameters
    ----------
    keys : int or tuple or list or array
        List of indices to slice an array.
    pos : int
        Index into keys.
    shape : int
        The length of the dimension we are slicing.

    Returns
    -------
    range
        The range.
    """
    length = key_len(keys)

    # Handle "slice the rest"
    if length > 1 and keys[-1] == Ellipsis:
        keys = keys[:-1]
        length -= 1

    if length > pos:
        try:
            val = keys[pos]
        except TypeError:
            # Only one key
            val = keys
        if isinstance(val, slice):
            if val.start is None and val.stop is None and val.step is None:
                return range(shape)
            tmp = range(10**10)[val]
            start, stop, step = tmp.start, tmp.stop, tmp.step
            if start > shape:
                start = shape
            if stop > shape:
                stop = shape
            return range(start, stop, step)
        elif isinstance(val, int):
            if val > shape or (val + 1) > shape:
                # TODO: It's not clear to me this is the correct behavior 
                return range(shape-1, shape)
            return range(val, val + 1)
            
    else:
        return range(shape)

def ensure_slice(keys, pos):
    """Ensure the output is a slice.

    Parameters
    ----------
    keys : int or tuple or list or array
        List of indices to slice an array.
    pos : int
        Index into keys.

    Returns
    -------
    slice
        The slice.
    """
    length = key_len(keys)

    # Handle "slice the rest"
    if length > 1 and keys[-1] == Ellipsis:
        keys = keys[:-1]
        length -= 1

    if length > pos:
        try:
            val = keys[pos]
        except TypeError:
            # Only one key
            val = keys
        if isinstance(val, int):
            return slice(val, val+1, None)
        assert isinstance(val, slice)
        return val
    else:
        # Default to all values
        return slice(None, None, None)


camera_device_types = {
    "Hamamatsu ORCA Lightning": ("HamamatsuOrcaLightning", "hamamatsu"),
    "Hamamatsu ORCA Fire": ("HamamatsuOrcaFire", "hamamatsu"),
    "Hamamatsu ORCA Fusion": ("HamamatsuOrcaFusion", "hamamatsu"),
    "Hamamatsu Flash 4.0": ("HamamatsuOrca", "hamamatsu"),
    "Photometrics Iris 15B": ("Photometrics", "photometrics"),
    "Virtual Device": ("Synthetic", "synthetic"),
}

camera_hardware_widgets = {
    "hardware/type": ["Device Type", "Combobox", "string", camera_device_types, None],
    "hardware/serial_number": [
        "Serial Number",
        "Input",
        "string",
        None,
        'Example: "302352"',
    ],
    "hardware/camera_connection": [
        "Camera Connection",
        "Input",
        "string",
        None,
        "*Photometrics Iris 15B only",
    ],
    "defect_correct_mode": [
        "Defect Correct Mode",
        "Combobox",
        "string",
        {"On": 2.0, "Off": 1.0},
        None,
    ],
    "delay": [
        "Delay (ms)",
        "Spinbox",
        "float",
        {"from": 0, "to": 100, "step": 0.1},
        None,
    ],
    "settle_down": [
        "Settle Down (ms)",
        "Spinbox",
        "float",
        {"from": 0, "to": 100, "step": 0.1},
        None,
    ],
    "flip_x": ["Flip X", "Checkbutton", "bool", None, None],
    "flip_y": ["Flip Y", "Checkbutton", "bool", None, None],
    "supported_channel_count": [
        "Microscope Channel Count",
        "Spinbox",
        "int",
        {"from": 5, "to": 10, "step": 1},
        None,
    ],
}

filter_wheel_device_types = {
    "Sutter Instruments": ("Sutter", "sutter"),
    "ASI Filter Wheel": ("ASI", "asi"),
    "ASI Dichroic Slider": ("ASICubeSlider", "asi"),
    "Ludl Electronic Products": ("LUDL", "ludl"),
    "Analog/Digital Device": ("NI", "ni"),
    "Virtual Device": ("Synthetic", "synthetic"),
}

filter_wheel_widgets = {
    "filter_name": ["Filter Name", "Input", "string", None, "Example: Empty-Alignment"],
    "filter_value": ["Filter Value", "Input", "string", None, "Example: 0"],
    "button_1": ["Delete", "Button", {"delete": True}],
    "frame_config": {
        "ref": "available_filters",
        "format": "item(filter_name,filter_value),",
        "direction": "horizon",
    },
}

filter_wheel_hardware_widgets = {
    "hardware/type": [
        "Device Type",
        "Combobox",
        "string",
        filter_wheel_device_types,
        None,
    ],
    "hardware/wheel_number": ["Number of Wheels", "Spinbox", "int", None, "Example: 1"],
    "hardware/port": ["Serial Port", "Input", "string", None, "Example: COM1"],
    "hardware/baudrate": ["Baudrate", "Input", "int", None, "Example: 9600"],
    "hardware/name": ["GUI Label", "Input", "string", None, "Example: Filter Wheel 1"],
    "filter_wheel_delay": [
        "Filter Wheel Delay (s)",
        "Input",
        "float",
        None,
        "Example: 0.03",
    ],
    "button_1": [
        "Add Available Filters",
        "Button",
        {
            "widgets": filter_wheel_widgets,
            "ref": "available_filters",
            "direction": "horizon",
            "parent": "constants",
        },
    ],
}

daq_device_types = {
    "National Instruments": "NI",
    "Applied Scientific Instrumentation": "ASI",
    "Virtual Device": "Synthetic",
}

daq_hardware_widgets = {
    "hardware/type": ["Device Type", "Combobox", "string", daq_device_types, None],
    "sample_rate": ["Sample Rate", "Input", "int", None, "Example: 100000"],
    "master_trigger_out_line": [
        "Master Trigger Out",
        "Input",
        "string",
        None,
        "Example: PXI6259/port0/line1",
    ],
    "camera_trigger_out_line": [
        "Camera Trigger Out",
        "Input",
        "string",
        None,
        "Example: /PXI6259/ctr0",
    ],
    "trigger_source": [
        "Trigger Source",
        "Input",
        "string",
        None,
        "Example: /PXI6259/PFI0",
    ],
    "laser_port_switcher": [
        "Laser Switcher Port",
        "Input",
        "string",
        None,
        "Example: PXI6733/port0/line0",
    ],
    "laser_switch_state": [
        "Laser Switch On State",
        "Combobox",
        "bool",
        [True, False],
        None,
    ],
}

shutter_device_types = {
    "ASI Shutter": ("ASI", "asi"),
    "Analog/Digital Device": ("NI", "ni"),
    "Virtual Device": ("Synthetic", "synthetic"),
}

shutter_hardware_widgets = {
    "type": ["Device Type", "Combobox", "string", shutter_device_types, None],
    "channel": ["NI Channel", "Input", "string", None, "Example: PXI6259/port0/line0"],
    "port": ["COM Port", "Input", "string", None, "Example: COM3"],
    "axis": ["Shutter Axis", "Input", "string", None, "Example: 1"],
    "min": ["Minimum Voltage", "Spinbox", "float", None, "Example: 0"],
    "max": ["Maximum Voltage", "Spinbox", "float", None, "Example: 5"],
    "frame_config": {"ref": "hardware"},
}

stage_device_types = {
    "Applied Scientific Instrumentation": ("ASI", "asi"),
    "ASI MFC2000": ("MFC2000", "asi"),
    "ASI MS2000": ("MS2000", "asi"),
    "Analog/Digital Device": ("NI", "ni"),
    "Mad City Labs": ("MCL", "mcl"),
    "Physik Instrumente": ("PI", "pi"),
    "Sutter Instruments": ("MP285", "sutter"),
    "ThorLabs KCube Inertial Device KIM001": ("KIM001", "thorlabs"),
    "ThorLabs KCube Inertial Device KST101": ("KST101", "thorlabs"),
    "Virtual Device": ("Synthetic", "synthetic"),
}

stage_hardware_widgets = {
    "type": ["Device Type", "Combobox", "string", stage_device_types, None],
    "serial_number": ["Serial Number", "Input", "string", None, None],
    "axes": [
        "Axes",
        "Input",
        "string",
        None,
        "Example: [x, y, z, theta, f]",
        "[x, y, z]",
    ],
    "axes_mapping": [
        "Axes Mapping",
        "Input",
        "string",
        None,
        "Example: [X, M, Y, D, E]",
        "[X, M, Y]",
    ],
    "feedback_alignment": [
        "Feedback Alighment",
        "Input",
        "string",
        None,
        "*ASI stage only. Example: [90, 90, 90, 0, 90]",
        "[90, 90, 90]",
    ],
    "device_units_per_mm": [
        "Device Units Per Micron",
        "Input",
        "float",
        None,
        "*KST101 only. Example: 2000.0",
        1000.25,
    ],
    "volts_per_micron": [
        "Volts Per Micron",
        "Input",
        "string",
        None,
        "*Analog/Digital Device only. Example: '0.1*x+0.05'",
        "0.1*x+0.05",
    ],
    "min": [
        "Minimum Volts",
        "Spinbox",
        "float",
        {"from": 0, "to": 5, "step": 0.1},
        "*Analog/Digital Device only",
        0,
    ],
    "max": [
        "Maximum Volts",
        "Spinbox",
        "float",
        {"from": 1, "to": 100, "step": 0.1},
        "*Analog/Digital Device only",
        5,
    ],
    "distance_threshold": [
        "Distance Threshold",
        "Spinbox",
        "float",
        {"from": 0, "to": 100, "step": 1},
        "*Analog-Controlled Galvo/Peizo only",
        5,
    ],
    "settle_duration_ms": [
        "Settle Duration (ms)",
        "Spinbox",
        "float",
        {"from": 0, "to": 100, "step": 1},
        "*Analog-Controlled Galvo/Peizo only",
        20,
    ],
    "controllername": [
        "Controller Name",
        "Input",
        "string",
        None,
        "*Physik Instrumente only. Example: 'C-884'",
    ],
    "stages": [
        "PI Stages",
        "Input",
        "string",
        None,
        "*Physik Instrumente only. Example: L-509.20DG10 L-509.40DG10",
    ],
    "refmode": [
        "REF Modes",
        "Input",
        "string",
        None,
        "*Physik Instrumente only. Example: FRF FRF",
    ],
    "port": ["Serial Port", "Input", "string", None, "Example: COM1"],
    "baudrate": ["Baudrate", "Input", "int", None, "Example: 9600"],
    "timeout": ["Serial Timeout", "Input", "float", None, "Example: 0.25", 0.25],
    "button_2": ["Delete", "Button", {"delete": True}],
    "frame_config": {
        "collapsible": True,
        "title": "Stage",
        "ref": "hardware",
        "format": "list-dict",
    },
}

stage_top_widgets = {
    "button_1": [
        "Add New Stage Device",
        "Button",
        {"widgets": stage_hardware_widgets, "ref": "hardware", "parent": "hardware"},
    ],
}

stage_constants_widgets = {
    "joystick_axes": [
        "Joystick Axes",
        "Input",
        "string",
        None,
        "Example: [x, y, z]",
        "[x, y, z]",
    ],
    "x_min": [
        "Min X",
        "Spinbox",
        "float",
        {"from": -100000, "to": 100000, "step": 1000},
        None,
        -10000,
    ],
    "x_max": [
        "Max X",
        "Spinbox",
        "float",
        {"from": 0, "to": 100000, "step": 1000},
        None,
        10000,
    ],
    "y_min": [
        "Min Y",
        "Spinbox",
        "float",
        {"from": -100000, "to": 100000, "step": 1000},
        None,
        -10000,
    ],
    "y_max": [
        "Max Y",
        "Spinbox",
        "float",
        {"from": 0, "to": 100000, "step": 1000},
        None,
        10000,
    ],
    "z_min": [
        "Min Z",
        "Spinbox",
        "float",
        {"from": -100000, "to": 10000, "step": 1000},
        None,
        -10000,
    ],
    "z_max": [
        "Max Z",
        "Spinbox",
        "float",
        {"from": 0, "to": 100000, "step": 1000},
        None,
        10000,
    ],
    "theta_min": [
        "Min Theta",
        "Spinbox",
        "float",
        {"from": 0, "to": 360, "step": 1},
        None,
        0,
    ],
    "theta_max": [
        "Max Theta",
        "Spinbox",
        "float",
        {"from": 0, "to": 360, "step": 1},
        None,
        360,
    ],
    "f_min": [
        "Min Focus",
        "Spinbox",
        "float",
        {"from": -100000, "to": 100000, "step": 1000},
        None,
        -10000,
    ],
    "f_max": [
        "Max Focus",
        "Spinbox",
        "float",
        {"from": 0, "to": 100000, "step": 1000},
        None,
        10000,
    ],
    "x_offset": [
        "Offset of X",
        "Spinbox",
        "float",
        {"from": -10000, "to": 10000, "step": 1000},
        None,
        0,
    ],
    "y_offset": [
        "Offset of Y",
        "Spinbox",
        "float",
        {"from": -10000, "to": 10000, "step": 100},
        None,
        0,
    ],
    "z_offset": [
        "Offset of Z",
        "Spinbox",
        "float",
        {"from": -10000, "to": 10000, "step": 10},
        None,
        0,
    ],
    "theta_offset": [
        "Offset of Theta",
        "Spinbox",
        "float",
        {"from": 0, "to": 360, "step": 1},
        None,
        0,
    ],
    "f_offset": [
        "Offset of Focus",
        "Spinbox",
        "float",
        {"from": -10000, "to": 10000, "step": 10},
        None,
        0,
    ],
    "flip_x": ["Flip X", "Checkbutton", "bool", None, None],
    "flip_y": ["Flip Y", "Checkbutton", "bool", None, None],
    "flip_z": ["Flip Z", "Checkbutton", "bool", None, None],
    "flip_f": ["Flip F", "Checkbutton", "bool", None, None],
    "frame_config": {"collapsible": True, "title": "Stage Constants"},
}

remote_focus_device_types = {
    "Equipment Solutions": ("EquipmentSolutions", "equipment_solutions"),
    "Analog Device": ("NI", "ni"),
    "ASI Device": ("ASI", "asi"),
    "Virtual Device": ("Synthetic", "synthetic"),
}

remote_focus_hardware_widgets = {
    "type": ["Device Type", "Combobox", "string", remote_focus_device_types, None],
    "channel": ["DAQ Channel", "Input", "string", None, "Example: PXI6259/ao3"],
    "axis": ["Device Type", "Input", "string", None, "Example: A"],
    "min": [
        "Minimum Voltage",
        "Spinbox",
        "float",
        {"from": -10, "to": 10, "step": 1},
        None,
    ],
    "max": [
        "Maximum Voltage",
        "Spinbox",
        "float",
        {"from": 0, "to": 10, "step": 1},
        None,
    ],
    "port": ["Serial Port", "Input", "string", None, "*Equipment Solutions only"],
    "baudrate": [
        "Baudrate",
        "Input",
        "int",
        None,
        "*Equipment Solutions only. Example: 9600",
    ],
    "frame_config": {"ref": "hardware"},
}

galvo_device_types = {"Analog Device": ("NI", "ni"), "ASI Device": ("ASI", "asi"), "Virtual Device": ("Synthetic", "synthetic")}

waveform_types = {
    "Sine": "sine",
    "Sawtooth": "sawtooth",
    "Square": "square",
}

galvo_hardware_widgets = {
    "hardware/type": ["Device Type", "Combobox", "string", galvo_device_types, None],
    "hardware/channel": [
        "DAQ Channel",
        "Input",
        "string",
        None,
        "*Analog Device only. Example: PXI6259/ao1",
    ],
    "hardware/axis": ["Axis", "Input", "string", None, "Example: A"],
    "hardware/min": [
        "Minimum Voltage",
        "Spinbox",
        "float",
        {"from": -10, "to": 10, "step": 0.1},
        None,
    ],
    "hardware/max": [
        "Maximum Voltage",
        "Spinbox",
        "float",
        {"from": 0, "to": 10, "step": 0.1},
        None,
    ],
    "waveform": ["Waveform", "Combobox", "string", waveform_types, None],
    "phase": [
        "Phase",
        "Spinbox",
        "float",
        {"from": 0, "to": 10, "step": 0.1},
        "Example: 1.57",
        0,
    ],
    "button_1": ["Delete", "Button", {"delete": True}],
    "frame_config": {
        "collapsible": True,
        "title": "Galvo Device",
        "ref": "None",
        "format": "list-dict",
    },
}

galvo_top_widgets = {
    "button_1": [
        "Add New Device",
        "Button",
        {"widgets": galvo_hardware_widgets, "parent": "hardware"},
    ],
}

zoom_device_types = {"Dynamixel": ("Dynamixel", "dynamixel"), "Virtual Device": ("Synthetic", "synthetic")}

zoom_position_widgets = {
    "zoom_value": ["Zoom Value", "Input", "string", None, "Example: 16x"],
    "position": ["Position", "Input", "float", None, "Example: 1000"],
    "pixel_size": ["Pixel Size (um)", "Input", "float", None, "Example: 0.5"],
    "button_1": ["Delete", "Button", {"delete": True}],
    "frame_config": {
        "ref": "position;pixel_size",
        "format": "item(zoom_value, position);item(zoom_value, pixel_size)",
        "direction": "horizon",
    },
}

zoom_hardware_widgets = {
    "type": ["Device Type", "Combobox", "string", zoom_device_types, None],
    "servo_id": ["Servo ID", "Input", "string", None, "Example: 1"],
    "port": ["Serial Port", "Input", "string", None, "Example: COM1"],
    "baudrate": ["Baudrate", "Input", "int", None, "Example: 9600"],
    "button_1": [
        "Add Zoom Value",
        "Button",
        {
            "widgets": zoom_position_widgets,
            "ref": "position;pixel_size",
            "direction": "horizon",
            "parent": "constants",
        },
    ],
    "frame_config": {"ref": "hardware"},
}

mirror_device_types = {
    "Imagine Optics": ("ImagineOptics", "imop"),
    "Virtual Device": ("Synthetic", "synthetic"),
}

mirror_hardware_widgets = {
    "hardware/type": ["Device Type", "Combobox", "string", mirror_device_types, None],
    "hardware/flat_path": [
        "Flat Mirror Path",
        "Input",
        "string",
        None,
        "Example: D:\WaveKitX64\MirrorFiles\Beads.wcs",
    ],
    "n_modes": ["Number of Modes", "Input", "int", None, "Example: 32", 32],
}

laser_device_types = {"Analog Device": ("NI", "ni"), "ASI Laser": ("ASI", "asi"), "Virtual Device": ("Synthetic", "synthetic")}

laser_hardware_widgets = {
    "wavelength": ["Wavelength", "Input", "int", None, "Example: 488", 488],
    "onoff": ["On/Off Setting", "Label", None, None, None],
    "onoff/hardware/type": ["Type", "Combobox", "string", laser_device_types, None],
    "onoff/hardware/axis": ["Digital Axis", "Input", "string", None, "Example: 2"],
    "onoff/hardware/channel": [
        "DAQ Channel",
        "Input",
        "string",
        None,
        "Example: PXI6733/port0/line2",
    ],
    "onoff/hardware/min": [
        "Minimum Voltage",
        "Spinbox",
        "float",
        {"from": 0, "to": 100, "step": 1},
        None,
    ],
    "onoff/hardware/max": [
        "Maximum Voltage",
        "Spinbox",
        "float",
        {"from": 0, "to": 100, "step": 1},
        None,
    ],
    "power": ["Power Setting", "Label", None, None, None],
    "power/hardware/type": ["Type", "Combobox", "string", laser_device_types, None],
    "power/hardware/axis": ["Analog Axis", "Input", "string", None, "Example: B"],
    "power/hardware/channel": [
        "DAQ Channel",
        "Input",
        "string",
        None,
        "Example: PXI6733/ao0",
    ],
    "power/hardware/min": [
        "Minimum Voltage",
        "Spinbox",
        "float",
        {"from": 0, "to": 1000, "step": 1},
        None,
    ],
    "power/hardware/max": [
        "Maximum Voltage",
        "Spinbox",
        "float",
        {"from": 0, "to": 1000, "step": 1},
        None,
    ],
    "button_1": ["Delete", "Button", {"delete": True}],
    "frame_config": {
        "collapsible": True,
        "title": "Wavelength",
        "format": "list-dict",
        "ref": "None",
    },
}

laser_top_widgets = {
    "button_1": [
        "Add Wavelength",
        "Button",
        {"widgets": laser_hardware_widgets, "parent": "hardware"},
    ],
}

hardwares_dict = {
    "Camera": camera_hardware_widgets,
    "Data Acquisition Card": daq_hardware_widgets,
    "Filter Wheel": (None, filter_wheel_hardware_widgets, filter_wheel_widgets),
    "Galvo": (galvo_top_widgets, galvo_hardware_widgets, None),
    "Lasers": (laser_top_widgets, laser_hardware_widgets, None),
    "Remote Focus Devices": remote_focus_hardware_widgets,
    "Adaptive Optics": mirror_hardware_widgets,
    "Shutters": shutter_hardware_widgets,
    "Stages": (stage_top_widgets, stage_hardware_widgets, stage_constants_widgets),
    "Zoom Device": (None, zoom_hardware_widgets, zoom_position_widgets),
}

hardwares_config_name_dict = {
    "Camera": "camera",
    "Data Acquisition Card": "daq",
    "Filter Wheel": "filter_wheel",
    "Galvo": "galvo",
    "Lasers": "laser",
    "Remote Focus Devices": "remote_focus",
    "Adaptive Optics": "mirror",
    "Shutters": "shutter",
    "Stages": "stage",
    "Zoom Device": "zoom",
}

deceased_device_type_names = {
    "ASIMS2000": "MS2000",
    "ASIMFC2000": "MFC2000",
    "GalvoNIStage": "NI",
    "Thorlabs": "KIM001",
}


# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import unittest
from unittest.mock import patch
import subprocess

# Third Party Imports

# Local Imports
from navigate._commit import get_git_revision_hash, get_version_from_file


class TestGetGitRevisionHash(unittest.TestCase):
    def test_return_type(self):
        """Test that the function returns a string."""
        result = get_git_revision_hash()
        self.assertIsInstance(result, str)

    @patch("navigate._commit.subprocess.check_output")
    def test_if_not_git_repo(self, mock_check_output):
        mock_check_output.side_effect = [
            b"true",
            # Mock the return value for ["git", "rev-parse", "--is-inside-work-tree"]
            b"dummy_commit_hash"
            # Mock the return value for ["git", "rev-parse", "HEAD"]
        ]
        result = get_git_revision_hash()

        # Verify the correct calls were made to subprocess.check_output
        mock_check_output.assert_any_call(
            ["git", "rev-parse", "--is-inside-work-tree"], stderr=subprocess.DEVNULL
        )
        mock_check_output.assert_any_call(["git", "rev-parse", "HEAD"])

        # Assert the returned commit hash
        self.assertEqual(result, "dummy_commit_hash")

    @patch("navigate._commit.subprocess.check_output")
    def test_not_git_repo(self, mock_check_output):
        # throw subprocess.CalledProcessError
        mock_check_output.side_effect = subprocess.CalledProcessError(1, "git")

        result = get_git_revision_hash()

        # Assert that None is returned
        self.assertIsNone(result)


class TestGetVersionFromFile(unittest.TestCase):
    def test_file_found(self):
        result = get_version_from_file()
        self.assertIsInstance(result, str)
        self.assertIsNot(result, "unknown")

    @patch("builtins.open")
    def test_file_not_found(self, mock_open):
        mock_open.side_effect = FileNotFoundError
        result = get_version_from_file()
        self.assertEqual(result, "unknown")



# Copyright (c) 2021-2024  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


#  Standard Library Imports
import logging
from typing import Any, Dict

# Third Party Imports

# Local Imports
from navigate.model.devices.remote_focus.base import RemoteFocusBase
from navigate.model.devices.device_types import NIDevice
from navigate.tools.decorators import log_initialization

# # Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


@log_initialization
class NIRemoteFocus(RemoteFocusBase, NIDevice):
    """RemoteFocusNI Class - Analog control of the remote focus device."""

    def __init__(
        self,
        microscope_name: str,
        device_connection: Any,
        configuration: Dict[str, Any],
        *args,
        **kwargs,
    ) -> None:
        """Initialize the RemoteFocusNI class.

        Parameters
        ----------
        microscope_name : str
            The microscope name.
        device_connection : Any
            The device connection object.
        configuration : Dict[str, Any]
            The configuration dictionary.

        """
        super().__init__(microscope_name, device_connection, configuration)

        #: str: The trigger source (e.g., physical channel).
        self.trigger_source = configuration["configuration"]["microscopes"][
            microscope_name
        ]["daq"]["trigger_source"]

        #: object: The device connection object.
        self.daq = device_connection

        #: str: The board name.
        self.board_name = self.device_config["hardware"]["channel"].split("/")[0]

    def __del__(self):
        """Delete the RemoteFocusNI object.

        Deletion of the NIDAQ task is handled by the NIDAQ object."""
        pass

    def adjust(self, exposure_times, sweep_times, offset=None):
        """Adjust the waveform.

        This method adjusts the waveform.

        Parameters
        ----------
        exposure_times : dict
            Dictionary of exposure times for each selected channel
        sweep_times : dict
            Dictionary of sweep times for each selected channel
        offset : float
            The offset of the signal in volts.

        Returns
        -------
        waveform_dict : dict
            The waveform dictionary.
        """
        waveform_dict = super().adjust(exposure_times, sweep_times, offset)

        self.daq.analog_outputs[self.device_config["hardware"]["channel"]] = {
            "trigger_source": self.trigger_source,
            "waveform": waveform_dict,
        }

        return waveform_dict

    def move(self, exposure_times, sweep_times, offset=None):
        """Move the remote focus.

        This method moves the remote focus.

        Parameters
        ----------
        exposure_times : dict
            Dictionary of exposure times for each selected channel
        sweep_times : dict
            Dictionary of sweep times for each selected channel
        offset : float
            The offset of the signal in volts.
        """
        self.adjust(exposure_times, sweep_times, offset)
        self.daq.update_analog_task(self.board_name)
